#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/collector/ResourceConfigCollector.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.10.2019
# Last Modified Date:  02.01.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any

from ampel.config.collector.AbsDictConfigCollector import AbsDictConfigCollector
from ampel.log.AmpelLogger import VERBOSE


class ResourceConfigCollector(AbsDictConfigCollector):

	def add(self,
		arg: dict[str, Any],
		dist_name: str,
		version: str | float | int,
		register_file: str
	) -> None:
		""" """

		if not isinstance(arg, dict):
			self.error(
				f"Resource value must be a dict. "
				f"Offending value {arg}\n"
				f"{self.distrib_hint(register_file, dist_name)}"
			)
			return

		for k, v in arg.items():

			try:

				key = k

				# Global resource
				if k and k[0] == "%":
					key = k[1:]
					scope = "global"
				# Distribution scoped alias
				elif dist_name:
					key = f"{dist_name}/{k}"
					scope = "scoped"
				else:
					scope = ""

				if self.verbose:
					self.logger.log(VERBOSE,
						f"Adding {scope} resource '{k}' " +
						f"from file {register_file}" if register_file else ""
					)

				if self.check_duplicates(
					key, dist_name, version, register_file,
					section_detail = f"{scope} resource"
				):
					continue

				self.__setitem__(key, v)

			except Exception as e:
				self.error(
					f"Error occured while loading resource {k} " +
					self.distrib_hint(dist_name, register_file),
					exc_info = e
				)



import mongomock
import pytest

from ampel.t2.T2Worker import T2Worker
from ampel.t3.supply.select.T3FilteringStockSelector import T3FilteringStockSelector


@pytest.fixture
def processed_t2s(integration_context, ingest_tied_t2):
    assert (num_dps := integration_context.db.get_collection("t0").count_documents({}))
    t2 = T2Worker(
        context=integration_context, raise_exc=True, process_name="t2", run_dependent_t2s=True
    )

    num_docs = t2.run()
    t2 = integration_context.db.get_collection("t2")
    assert t2.count_documents({}) == num_docs

    if "point" in ingest_tied_t2.param.lower():
        assert num_docs == 1 + num_dps
        assert (
            next(t2.find({"unit": "DummyPointT2Unit"}).sort([("link", -1)]))["body"][0]["thing"] == 3
        )
        assert (
            t2.find_one({"unit": "DummyTiedStateT2Unit"})["body"][0]["thing"]
            == 2 * 3
        )
    elif "stock" in ingest_tied_t2.param.lower():
        assert num_docs == 2
        assert (
            t2.find_one({"unit": "DummyStockT2Unit"})["body"][0]["id"]
            == "stockystock"
        )
        assert (
            t2.find_one({"unit": "DummyTiedStateT2Unit"})["body"][0]["id"]
            == 2 * "stockystock"
        )
    else:
        assert num_docs == 2
        assert (
            t2.find_one({"unit": "DummyStateT2Unit"})["body"][0]["len"]
            == num_dps
        )
        assert (
            t2.find_one({"unit": "DummyTiedStateT2Unit"})["body"][0]["len"]
            == 2 * num_dps
        )

    return ingest_tied_t2.param


def test_filter(integration_context, processed_t2s, ampel_logger):
    if isinstance(integration_context.db.get_collection("t2").database.client, mongomock.MongoClient):
        pytest.skip("T3FilteringStockSelector does not work with mongomock")
    if processed_t2s != "DummyStateT2Unit":
        pytest.skip("Only deal with a single unit for now")
    selector = T3FilteringStockSelector(
        context=integration_context,
        logger=ampel_logger,
        t2_filter={"unit": processed_t2s, "match": {"len": {"$gt": 1}}},
    )

    assert len(ids := list(selector.fetch())) == 1
    assert ids[0] == {"stock": "stockystock"}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/ingest/T0Compiler.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                11.05.2021
# Last Modified Date:  25.11.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import datetime
from typing import Any

from ampel.abstract.AbsCompiler import AbsCompiler
from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.content.DataPoint import DataPoint
from ampel.content.MetaRecord import MetaRecord
from ampel.enum.MetaActionCode import MetaActionCode
from ampel.types import ChannelId, DataPointId
from ampel.util.collections import try_reduce


class T0Compiler(AbsCompiler):
	"""
	Compiles updates to t0 documents arising from different t0 directives
	"""

	def __init__(self, **kwargs) -> None:
		super().__init__(**kwargs)

		# We assume only the usage of only one shaper (that is only one tracid)
		self.register: dict[
			DataPointId,
			tuple[
				DataPoint,
				set[ChannelId],
				None | datetime.timedelta, # ttl
				None | int,           # trace id
				None | dict[str, Any] # meta extra
			]
		] = {}

		self.retained: dict[DataPointId, datetime.timedelta] = {}


	# Override
	def add(self, # type: ignore[override]
		dps: list[DataPoint],
		channel: ChannelId,
		ttl: None | datetime.timedelta,
		trace_id: None | int,
		extra: None | dict[str, Any] = None
	) -> None:

		r = self.register
		for dp in dps:
			dpid = dp['id']
			if dpid in r:
				r[dpid][1].add(channel)
				if ttl is not None and ((prev := r[dpid][2]) is None or ttl > prev):
					r[dpid] = r[dpid][:2] + (ttl,) + r[dpid][3:]
			else:
				r[dpid] = dp, {channel}, ttl, trace_id, extra


	def retain(self, dps: list[DataPoint], ttl: datetime.timedelta) -> None:
		for dp in dps:
			if dp['id'] not in self.retained or ttl > self.retained[dp['id']]:
					self.retained[dp['id']] = ttl

	# Override
	def commit(self, ingester: AbsDocIngester[DataPoint], now: int | float, **kwargs) -> None:
		"""
		Note that we let the ingester handle 'ts' and 'updated' values
		"""

		for dp, channel_sets, ttl, trace_id, extra in self.register.values():

			lchans = list(channel_sets)
			meta: MetaRecord = {'ts': now, 'run': self.run_id}

			if extra:
				meta.update(extra) # type: ignore[typeddict-item]

			meta['activity'] = [
				{
					'action': MetaActionCode.ADD_CHANNEL,
					'channel': try_reduce(lchans)
				}
			]
			meta['traceid'] = {'shaper': trace_id}

			if self._tag or dp.get('tag'):
				dp['tag'] = [*(self._tag or []), *(dp.get('tag') or [])]
				meta['activity'].append(self._ingest_tag_activity) # type: ignore[attr-defined]

			if self.origin and 'origin' not in dp:
				dp['origin'] = self.origin

			dp['channel'] = lchans
			if 'meta' in dp and isinstance(dp['meta'], list):
				dp['meta'].append(meta)
			else:
				dp['meta'] = [meta]

			if ttl is not None:
				dp['expiry'] = datetime.datetime.fromtimestamp(
					now, tz=datetime.timezone.utc
				) + ttl

			ingester.ingest(dp)

		# Retain datapoints that were not explicitly ingested
		if retained := set(self.retained).difference(self.register.keys()):
			nowdt = datetime.datetime.fromtimestamp(
				now, tz=datetime.timezone.utc
			)
			if hasattr(ingester, "update_expiry"):
				ingester.update_expiry({k: nowdt + self.retained[k] for k in retained})

		self.register.clear()
		self.retained.clear()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/utils.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                31.10.2018
# Last Modified Date:  09.10.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Sequence
from typing import Any

from ampel.types import StrictIterable, strict_iterable


def add_or(query: dict[str, Any], arg: dict[str, Any]) -> None:

	if '$or' in query:
		if '$and' in query:
			if not isinstance(query['$and'], list):
				raise ValueError(f"Illegal $and value in query: {query}")
			query['$and'].append(arg)
		else:
			query['$and'] = [query.pop('$or'), arg]

	else:
		query['$or'] = arg


# Rework function names later
def maybe_match_array(arg: StrictIterable):
	"""
	maybe_match_array(['ab']) -> returns 'ab'
	maybe_match_array({'ab'}) -> returns 'ab'
	maybe_match_array(['a', 'b']) -> returns {$in: ['a', 'b']}
	maybe_match_array({'a', 'b'}) -> returns {$in: ['a', 'b']}
	"""

	if not isinstance(arg, strict_iterable):
		raise ValueError(
			f"Provided argument is not sequence ({type(arg)})"
		)

	if len(arg) == 1:
		return next(iter(arg))

	if isinstance(arg, list):
		return {'$in': arg}

	# Otherwise cast to list
	return {'$in': list(arg)}


def match_one_or_many(arg: Any) -> Any:

	if isinstance(arg, int | float | str):
		return arg

	return maybe_match_array(arg)


def maybe_use_each(arg: Sequence[Any]) -> Any:
	if isinstance(arg, dict):
		return arg
	if len(arg) == 1:
		return next(iter(arg))
	return {'$each': arg}


def get_ids(col: Any, *, filter_stage: None | dict = None) -> set[Any]:
	"""
	Note1: timeit perf of find vs aggregate for cols with 10 / 700 elements:
	[el['_id'] for el in col.find()] -> 518 µs / 6.1 ms
	list(col.aggregate([{"$project": ..}, {"$group": ..}])) -> 536 µs / 1.72 ms

	Pagination will be required (using $sort, $skip, $limit) if the returned doc exceeds 16MB
	"""

	agg = [
		{"$project": {"_id": "$_id"}},
		{"$group": {"_id": None, "ids": {"$push": "$$ROOT._id"}}}
 	]

	if filter_stage:
		agg.insert(0, filter_stage)

	if res := next(col.aggregate(agg), None):
		return set(res['ids'])
	return set()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/collector/AbsListConfigCollector.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                03.03.2020
# Last Modified Date:  03.03.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any

from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.config.collector.ConfigCollector import ConfigCollector


class AbsListConfigCollector(ConfigCollector, AmpelABC, abstract=True):

	@abstractmethod
	def add(self,
		arg: list[Any],
		dist_name: str,
		version: str | float | int,
		register_file: str,
	) -> None | int:
		...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/export.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.08.2022
# Last Modified Date:  16.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import sys
from collections.abc import Generator
from datetime import datetime
from io import BufferedWriter, TextIOWrapper
from itertools import islice
from json import dumps
from typing import BinaryIO, TextIO

from bson import ObjectId, encode

from ampel.abstract.AbsIdMapper import AbsIdMapper
from ampel.protocol.LoggerProtocol import LoggerProtocol
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.util.getch import getch as fgetch
from ampel.util.pretty import prettyjson
from ampel.util.serialize import walk_and_encode

dsi = dict.__setitem__
ufts = datetime.utcfromtimestamp


def txt_export(
	fd: TextIOWrapper | TextIO,
	gen: Generator[AmpelBuffer, None, None],
	id_mapper: None | AbsIdMapper = None,
	chunk_size: int = 200,
	human_times: bool = True,
	pretty: bool = False,
	getch: bool = False,
	close_fd: bool = True,
	logger: None | LoggerProtocol = None
) -> None:

	first = True
	func = prettyjson if pretty else dumps

	fd.write('[\n')

	try:

		while (data := list(islice(gen, chunk_size))):

			walk_and_encode(data)

			if first and len(data) > 0:

				data = iter(data) # type: ignore[assignment]
				el = next(data) # type: ignore[call-overload]
				if id_mapper:
					el['id'] = id_mapper.to_ext_id(el['id'])
				if human_times:
					convert_timestamps(el)
				if logger:
					logger.info(f"Writing content (id: {el['id']})")
				fd.write(func(el))
				first = False

			if getch and fgetch():
				fd.write('\n]\n')
				if logger:
					logger.info("Abording")
				return

			for el in data:

				fd.write(",\n")
				if id_mapper:
					el['id'] = id_mapper.to_ext_id(el['id'])
				if human_times:
					convert_timestamps(el)
				if logger:
					logger.info(f"Writing content (id: {el['id']})")
				fd.write(func(el))

				if getch and fgetch():
					fd.write('\n]\n')
					if logger:
						logger.info("Abording")
					return

		fd.write('\n]\n')

	finally:

		if close_fd:
			fd.flush()
			fd.close()


def bin_export(
	fd: BufferedWriter | BinaryIO,
	gen: Generator[AmpelBuffer, None, None],
	id_mapper: None | AbsIdMapper = None,
	close_fd: bool = True,
	logger: None | LoggerProtocol = None
) -> None:

	for el in gen:
		if logger:
			logger.info(f"Writing content (id: {el['id']})") # type: ignore[str-bytes-safe]
		if id_mapper:
			el['id'] = id_mapper.to_ext_id(el['id'])
		fd.write(encode(el))

	if close_fd:
		if logger:
			logger.info("Closing file descriptor")
		fd.flush()
		fd.close()


def get_fd(
	binary: bool,
	fd: None | str | BufferedWriter | TextIO | TextIOWrapper | BinaryIO,
	close_fd: bool
) -> tuple[BufferedWriter | TextIO | TextIOWrapper | BinaryIO, bool]:

	if binary:
		if isinstance(fd, str):
			return open(fd, "wb"), True
		return fd, close_fd # type: ignore[return-value]
	if fd is None:
		return sys.stdout, False
	if isinstance(fd, str):
		return open(fd, 'w'), close_fd
	return fd, close_fd



def convert_timestamps(ab: AmpelBuffer) -> None:

	try:

		if stock_doc := ab.get('stock'):
			for k in (ts := stock_doc['ts']):
				dsi( # type: ignore[misc]
					ts, k, {
						'tied': ufts(ts[k]['tied']).isoformat(),
						'upd': ufts(ts[k]['upd']).isoformat()
					}
				)
			for je in stock_doc['journal']:
				dsi(je, 'ts', ufts(je['ts']).isoformat()) # type: ignore[arg-type]

		if t1s := ab.get('t1'):
			for t1doc in t1s:
				if isinstance(t1doc['_id'], str) and "oid:" in t1doc['_id']: # type: ignore[typeddict-item]
					gt = ObjectId(t1doc['_id'][4:]).generation_time # type: ignore[typeddict-item]
				else:
					gt = t1doc['_id'].generation_time # type: ignore[typeddict-item]
				dsi(t1doc, 'added', gt.isoformat()) # type: ignore[arg-type]
				for t1meta in t1doc['meta']:
					dsi(t1meta, 'ts', ufts(t1meta['ts']).isoformat()) # type: ignore[arg-type]

		if t2s := ab.get('t2'):
			for t2doc in t2s:
				for t2meta in t2doc['meta']:
					dsi(t2meta, 'ts', ufts(t2meta['ts']).isoformat()) # type: ignore[arg-type]

	except Exception:
		import sys
		import traceback
		traceback.print_exc()
		print(ab)
		sys.exit("Exception occured")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/ingest/T1Compiler.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                01.01.2018
# Last Modified Date:  21.11.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import datetime
from collections.abc import Sequence
from struct import pack
from typing import Any, NamedTuple

import xxhash

from ampel.abstract.AbsCompiler import AbsCompiler, ActivityRegister
from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.content.MetaActivity import MetaActivity
from ampel.content.T1Document import T1Document
from ampel.types import ChannelId, DataPointId, StockId, UBson, UnitId


class T1Compiler(AbsCompiler):
	"""
	Helps build a minimal set of :class:`T1Document <ampel.content.T1Document.T1Document>`
	that represent a collection of :class:`datapoints <ampel.content.DataPoint.DataPoint>`,
	as viewed through a set of channels.

	Different T1 units may select different subsets of datapoints associated with a stock.
	Only one :class:`~ampel.content.T1Document.T1Document` will be created for each
	effective subselection. This allows downstream calculations that operate on equivalent
	subselections to be performed only once.

	Excluded datapoints may be included as metadata in T1Document.meta.excl
	"""

	#: true: datapoints will have the same hash regardless of their order (unordered set)
	#: false: order of datapoints will influence the computed hash
	sort: bool = True

	#: Convention: negative size means the int hash result will be converted to signed int
	#: (necessary when using mongodb)
	#: Use 64/-64 if you expect more than 10K states per stock(s)
	hash_size: int = -32

	#: Change this only if you know what you're doing
	seed: int = 0

	class UnitKey(NamedTuple):
		unit: None | UnitId
		config: None | int
		stock: StockId
		dps: tuple[DataPointId, ...]

	class LinkTarget(NamedTuple):
		link: int
		channels: set[ChannelId]
		ttl: None | datetime.timedelta
		body: UBson
		code: None | int
		meta: dict[
			frozenset[tuple[str, Any]], # key: traceid
			tuple[
				ActivityRegister,
				dict[str, Any]  # meta extra
			]
		]


	def __init__(self, **kwargs) -> None:

		super().__init__(**kwargs)
		self.hasher = getattr(xxhash, f'xxh{abs(self.hash_size)}_intdigest')

		# Internal structure used for compiling documents
		self.t1s: dict[T1Compiler.UnitKey, T1Compiler.LinkTarget] = {}


	def add(self, # type: ignore[override]
		dps: Sequence[DataPointId],
		channel: ChannelId,
		ttl: None | datetime.timedelta,
		traceid: dict[str, Any],
		stock: StockId = 0,
		activity: None | MetaActivity | list[MetaActivity] = None,
		meta_extra: None | dict[str, Any] = None,
		unit: None | UnitId = None,
		config: None | int = None,
		body: UBson = None,
		code: None | int = None
	) -> int:
		"""
		:param unit: potential AbsT1ComputeUnit subclass to be associated with this doc
		:param config: config of the AbsT1ComputeUnit subclass associated with this doc
		"""

		if not dps:
			raise ValueError("Argument dps cannot be empty")

		if self.sort:
			# Note: int DataPointId is implicitely required,
			# hashing of single dp ids will be required if bytes is used rather than int
			dps = sorted(dps)

		k = T1Compiler.UnitKey(unit, config, stock, tuple(dps))
		tid = frozenset(traceid.items())

		# a: (link, {channels}, body, code, dict[traceid, (ActivityRegister, meta extra)])
		# a: ( 0  ,     1     ,  2  ,  3  ,                   4                        )])
		if a := self.t1s.get(k):

			a.channels.add(channel)

			# One meta record will be created for each unique trace id
			# (the present compiler could be used by multiple handlers configured differently)
			# activity can contain channel-based dps exclusions (excl: [1232, 3232])
			if tid in a.meta:

				# (activity register, meta_extra)
				b = a.meta[tid]

				# update internal register
				self.register_meta_info(b[0], b[1], channel, activity, meta_extra)
			else:
				a.meta[tid] = self.new_meta_info(channel, activity, meta_extra)

			# Update ttl if necessary
			if ttl is not None and (a.ttl is None or ttl > a.ttl):
				self.t1s[k] = T1Compiler.LinkTarget(a.link, a.channels, ttl, a.body, a.code, a.meta)

		else:

			i = self.hasher(
				pack(f">{len(dps)}q", *dps),
				self.seed
			)

			# Convert unsigned to signed
			if self.hash_size < 0 and i & (1 << (-self.hash_size-1)):
				i -= 2**-self.hash_size

			a = self.t1s[k] = T1Compiler.LinkTarget(
				i, {channel}, ttl, body, code,
				{tid: self.new_meta_info(channel, activity, meta_extra)}
			)

		return a.link


	def commit(self, ingester: AbsDocIngester[T1Document], now: int | float, **kwargs) -> None:

		# t1: (unit, config, stock, dps)
		# t2: (link, {channels}, body, code, dict[traceid, (ActivityRegister, meta extra)])
		# t2: ( 0  ,     1     ,  2  ,  3  ,                   4                        )])
		for t1, t2 in self.t1s.items():

			# Note: mongodb maintains key order
			d: T1Document = {'link': t2.link} # type: ignore[typeddict-item]

			if t1.unit:
				d['unit'] = t1.unit

			if t1.config is not None:
				d['config'] = t1.config

			d['stock'] = t1.stock

			if self.origin:
				d['origin'] = self.origin

			d['channel'] = list(t2.channels)
			d['dps'] = list(t1.dps)

			d['meta'], tags = self.build_meta(t2.meta, now)

			if tags:
				d['tag'] = tags

			if t2.body is not None:
				d['body'] = [t2.body]

			if t2.code is not None:
				d['code'] = t2.code

			if t2.ttl is not None:
				d['expiry'] = datetime.datetime.fromtimestamp(
					now, tz=datetime.timezone.utc
				) + t2.ttl

			ingester.ingest(d)

		self.t1s.clear()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/builder/ConfigChecker.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                03.09.2019
# Last Modified Date:  26.02.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>


import json
from collections.abc import Generator
from typing import Any

from ampel.config.AmpelConfig import AmpelConfig
from ampel.core.UnitLoader import UnitLoader
from ampel.log.AmpelLogger import DEBUG, ERROR, AmpelLogger
from ampel.secret.AmpelVault import AmpelVault
from ampel.secret.PotemkinSecretProvider import PotemkinSecretProvider


class BaseConfigChecker:
	"""
	Validates a config usually build by ConfigBuilder
	"""

	def __init__(self, config: dict[str, Any], logger: None | AmpelLogger = None, verbose: bool = False):

		self.verbose = verbose
		self.logger = AmpelLogger.get_logger(
			console={'level': DEBUG if verbose else ERROR}
		) if logger is None else logger

		self.loader = UnitLoader(
			AmpelConfig(config),
			db=None,
			provenance=False,
			vault=AmpelVault([PotemkinSecretProvider()])
		)

		# Fast deep copy and serialization check
		self.config = json.loads(json.dumps(config))


	def iter_procs(self,
		ignore_inactive: bool = False,
		raise_exc: bool = False
	) -> Generator[tuple[str, str], None, None]:

		for tier in ("t0", "t1", "t2", "t3"):

			procs = list(self.config['process'][tier].keys())

			for proc in procs:

				if ignore_inactive and not self.config['process'][tier][proc].get('active', True):
					self.logger.info("Ignoring inactivated processor model", extra={"process": proc})
					continue

				yield (tier, proc)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/view/MongoOrView.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                26.03.2021
# Last Modified Date:  21.07.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any

from ampel.mongo.view.AbsMongoFlatMultiView import AbsMongoFlatMultiView


class MongoOrView(AbsMongoFlatMultiView):

	# Override to add other matching criteria (ex: tag-based selection)
	def get_first_match(self) -> dict[str, Any]:
		return {'$match': {'channel': {'$in': self.channel}}}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/core/AmpelContext.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                18.02.2020
# Last Modified Date:  09.01.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Literal

from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.config.AmpelConfig import AmpelConfig
from ampel.config.builder.DisplayOptions import DisplayOptions
from ampel.secret.AmpelVault import AmpelVault

# Avoid cyclic import issues
if TYPE_CHECKING:
	from ampel.core.AmpelDB import AmpelDB
	from ampel.core.UnitLoader import UnitLoader


class AmpelContext:
	"""
	This class is typically instantiated by Ampel controllers
	and provided to "processor unit" instances.
	Note: this class might in the future be capable
	of handling multiple AmpelConfig/AmpelDB instances
	"""

	def __init__(self,
		config: AmpelConfig,
		db: 'AmpelDB',
		loader: 'UnitLoader', # forward reference to avoid cyclic import issues
		resource: None | dict[str, Any] = None,
		admin_msg: None | str = None
	) -> None:
		"""
		:param config: system configuration
		:param db: database client
		:param loader: instantiates unit from :class:`~ampel.model.UnitModel.UnitModel`
		"""

		self.config = config
		self.db = db
		self.loader = loader
		self.admin_msg = admin_msg
		self.resource = resource
		self.run_time_aliases: dict[str, Any] = {}
		
		# try to register aux units globally
		try:
			AuxUnitRegister.initialize(config)
		except Exception:
			print("UnitLoader auxiliary units auto-registration failed")  # noqa: T201


	@classmethod
	def load(cls,
		config: str | dict,
		pwd_file_path: None | str = None,
		pwds: None | Iterable[str] = None,
		freeze_config: bool = True,
		vault: None | AmpelVault = None,
		one_db: bool = False,
		**kwargs
	) -> 'AmpelContext':
		"""
		Instantiates a new AmpelContext instance.

		:param config: local path to an ampel config file (yaml or json) or loaded config as dict
		:param pwd_file_path: path to a text file containing one password per line.
		The underlying AmpelVault will be initialized with an AESecretProvider configured with these pwds.
		:param pwds: Same as 'pwd_file_path' except a list of passwords is provided directly via this parameter.
		:param freeze_config:
			whether to convert the elements contained of ampel config
			into immutable structures (:class:`dict` ->
			:class:`~ampel.view.ReadOnlyDict.ReadOnlyDict`, :class:`list` -> :class:`tuple`).
			Parameter does only apply if the config is loaded by this method, i.e if parameter 'config' is a str.
		"""

		# Avoid cyclic import issues
		from ampel.core.UnitLoader import UnitLoader # noqa
		from ampel.core.AmpelDB import AmpelDB

		alconf = AmpelConfig(config) if isinstance(config, dict) else AmpelConfig.load(config)
		if vault is None:
			vault = AmpelVault([])

		if pwd_file_path and not pwds:
			with open(pwd_file_path) as f:
				pwds = [l.strip() for l in f.readlines()]

		if pwds:
			# AESecretProvider is optional
			from ampel.secret.AESecretProvider import AESecretProvider
			vault.providers.append(
				AESecretProvider(pwds)
			)

		db = AmpelDB.new(alconf, vault, one_db=one_db)

		return cls(
			config = alconf,
			db = db,
			loader = UnitLoader(config=alconf, db=db, vault=vault),
			**kwargs
		)


	@classmethod
	def build(cls,
		ignore_errors: bool = False,
		pwd_file_path: None | str = None,
		pwds: None | Iterable[str] = None,
		freeze_config: bool = True,
		verbose: bool = False,
	) -> 'AmpelContext':
		"""
		Instantiates a new AmpelContext instance.

		The required underlying ampel configuration (:class:`~ampel.config.AmpelConfig.AmpelConfig`) is built from scratch,
		meaning all available configurations defined in the various ampel repositories available locally
		are collected, merged merged together and morphed the info into the final ampel config.
		This is a convenience method, do not use for production
		"""

		if pwd_file_path:
			with open(pwd_file_path) as f:
				pwds = [l.strip() for l in f.readlines()]

		# Import here to avoid cyclic import error
		from ampel.config.builder.DistConfigBuilder import DistConfigBuilder
		cb = DistConfigBuilder(options=DisplayOptions(verbose=verbose))
		cb.load_distributions()

		return cls.load(
			cb.build_config(ignore_errors, pwds=pwds),
			pwd_file_path=pwd_file_path,
			pwds=pwds,
			freeze_config=freeze_config,
		)


	def new_run_id(self) -> int:
		"""
		Return an identifier that can be used to associate log entries from a
		single process invocation. This ID is unique and monotonicaly increasing.
		"""
		return self.db \
			.get_collection('counter') \
			.find_one_and_update(
				{'_id': 'current_run_id'},
				{'$inc': {'value': 1}},
				new=True, upsert=True
			) \
			.get('value')


	def get_config(self) -> AmpelConfig:
		"""
		.. note:: in the future, AmpelContext might hold references to multiple different config
		"""
		return self.config


	def get_database(self) -> 'AmpelDB':
		"""
		.. note:: in the future, this class might hold references to multiple different databases
		"""
		return self.db


	def resolve_conf_id(self, conf_id: int) -> None | dict[str, Any]:
		try:
			return self.config.get_conf_id(conf_id)
		# confid not found (obsolete or dynamically generated by process)
		except Exception:
			l = list(self.db.col_conf_ids.find({"_id": conf_id}))
			if len(l) == 0:
				return None
			del l[0]['_id']
			return l[0]


	def add_conf_id(self, conf_id: int, unit_config: dict[str, Any]) -> None:
		self.db.add_conf_id(conf_id, unit_config)
		dict.__setitem__(self.config._config["confid"], conf_id, unit_config)  # noqa: SLF001


	def add_run_time_alias(self, key: str, value: Any, overwrite: bool = False) -> None:
		if not isinstance(key, str) or key[0] != '%' != key[1]:
			raise ValueError('Run time aliases must begin with %%')
		if key in self.run_time_aliases and not overwrite:
			raise ValueError(f"Run time alias {key} already defined, set overwrite=True to ignore")
		self.run_time_aliases[key] = value


	def __repr__(self) -> str:
		return "<AmpelContext>"


	def deactivate_processes(self) -> None:
		for i in range(4):
			for p in self.config.get(f'process.t{i}', dict, raise_exc=True).values():
				p['active'] = False


	def activate_process(self,
		name: str,
		tier: None | Literal[0, 1, 2, 3] = None
	) -> bool:
		"""
		:returns: False if process was deactivated, True if not found
		"""
		for i in range(4):
			if tier and i != tier:
				continue
			if p := self.config.get(f'process.t{i}.{name}', dict):
				p['active'] = False
				return False
		return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/ops/AmpelExceptionPublisher.py
# License:             BSD-3-Clause
# Author:              Jakob van Santen <jakob.van.santen@desy.de>
# Date:                03.09.2018
# Last Modified Date:  27.08.2020
# Last Modified By:    Jakob van Santen <jakob.van.santen@desy.de>


import datetime
import json
import socket
from typing import Any

from bson import ObjectId
from slack_sdk.web import WebClient
from slack_sdk.web.slack_response import SlackResponse

from ampel.abstract.AbsOpsUnit import AbsOpsUnit
from ampel.secret.NamedSecret import NamedSecret


class AmpelExceptionPublisher(AbsOpsUnit):

    slack_token: NamedSecret[str] = NamedSecret[str](label="slack/operator")
    user: str = f"ampel@{socket.gethostname()}"
    channel: str = "ampel-troubles"
    dry_run: bool = False
    quiet: bool = True

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.slack = WebClient(self.slack_token.get())
        self.troubles = self.context.db.get_collection("trouble", "r")

    def t3_fields(self, doc: dict[str, Any]) -> list[dict[str, Any]]:
        fields = []
        if "job" in doc:
            fields.append(
                {"title": "Job", "value": doc.get("job"), "short": True}
            )
        if "task" in doc:
            fields.append(
                {"title": "Task", "value": doc.get("task"), "short": True}
            )
        fields.append({"title": "Run", "value": doc.get("run"), "short": True})
        return fields

    def format_attachment(self, doc: dict[str, Any]) -> dict[str, Any]:
        fields = [
            {"title": "tier", "value": doc["tier"], "short": True},
        ]
        if "process" in doc:
            fields.append({"title": "process", "value": doc["process"], "short": True})
        if "run" in doc:
            fields.append({"title": "run", "value": doc["run"], "short": True})
        if doc["tier"] == 0:
            for field in "section", "stock":
                fields.append(  # noqa: PERF401
                    {"title": field, "value": doc.get(field), "short": True}
                )
            if "alert" in doc:
                fields.append(
                    {"title": "alert", "value": doc["alert"], "short": True}
                )
        elif doc["tier"] == 2:
            fields.append(
                {"title": "unit", "value": doc.get("unit"), "short": True}
            )
            t2Doc = doc.get("t2Doc")
            if t2Doc and hasattr(t2Doc, "binary"):
                fields.append(
                    {"title": "t2Doc", "value": t2Doc.binary.hex(), "short": True}
                )
        if "exception" in doc:
            text = "```{}```".format("\n".join(doc["exception"]))
        elif "location" in doc:
            text = "{}: {}".format(doc["location"], doc.get("ampelMsg", ""))
            if "mongoUpdateResult" in doc:
                text += "\nmongoUpdateResult: `{}`".format(doc["mongoUpdateResult"])
            elif "errDict" in doc:
                text += "```\n{}```".format(repr(doc["errDict"]))
        else:
            text = f"Unknown exception type. Doc keys are: ```{doc.keys()}```"

        return {
            "fields": fields,
            "ts": int(doc["_id"].generation_time.timestamp()),
            "text": text,
            "mrkdwn_in": ["text"],
        }

    def run(self, beacon: None | dict[str, Any] = None) -> None | dict[str, Any]:

        now = datetime.datetime.now(tz=datetime.timezone.utc)
        t0 = beacon["updated"] if beacon else now - datetime.timedelta(hours=1)
        dt = now - t0

        attachments: list[dict[str, Any]] = []
        message = {
            "attachments": attachments,
            "channel": "#" + self.channel,
            "username": self.user,
            "as_user": False,
        }

        docs = list(self.troubles.find({"_id": {"$gt": ObjectId.from_datetime(t0)}}))
        for doc in docs:
            if len(attachments) < 20:
                attachments.append(self.format_attachment(doc))
            else:
                break

        if dt.days > 3:
            time_range = f"{dt.days} days"
        elif dt.days > 0 or dt.seconds > 2 * 3600:
            time_range = f"{int(dt.days * 24 + dt.seconds / 3600)} hours"
        elif dt.seconds > 2 * 60:
            time_range = f"{int(dt.seconds / 60)} minutes"
        else:
            time_range = f"{int(dt.seconds)} seconds"

        count = len(docs)
        if len(attachments) < count:
            message[
                "text"
            ] = f"Here are the first {len(attachments)} exceptions. There were {count-len(attachments)} more in the last {time_range}."
        else:
            message[
                "text"
            ] = f"There were {len(attachments)} exceptions in the last {time_range}."

        if self.dry_run:
            self.logger.info(json.dumps(message, indent=1))
        elif attachments or not self.quiet:
            result = self.slack.api_call("chat.postMessage", json=message)
            if isinstance(result, SlackResponse):
                if not result["ok"]:
                    raise RuntimeError(result["error"])
            else:
                raise TypeError(f"Sync client returned a future {result}")
        self.logger.info(f"{count} exceptions in the last {time_range}".format(count))

        return {"updated": now}



import sys
from collections.abc import Sequence
from contextlib import contextmanager
from pathlib import Path

import pytest
import yaml
from mongomock import MongoClient
from pytest_mock import MockerFixture

from ampel.cli.main import main
from ampel.cli.T2Command import T2Command
from ampel.core.AmpelDB import UnknownDatabase


@contextmanager
def argv_context(args: Sequence[str]):
    argv = sys.argv
    try:
        sys.argv = list(args)
        yield
    finally:
        sys.argv = argv


def run(args: Sequence[str]) -> None | int | str:
    try:
        with argv_context(args):
            main()
        return None
    except SystemExit as se:
        return se.code


def dump(payload, tmpdir, name: str) -> Path:
    f = Path(tmpdir) / name
    with f.open("w") as fd:
        yaml.dump(payload, fd)
    return f


@pytest.fixture
def secrets():
    return {
        "mongo/writer": {"username": "writer", "password": "writer"},
        "mongo/logger": {"username": "logger", "password": "logger"},
    }


@pytest.fixture
def vault(tmpdir, secrets):
    return dump(secrets, tmpdir, "secrets.yml")


@pytest.fixture
def check_mongo_auth(mocker):
    def checky(*args, **kwargs):
        assert {"username", "password"}.issubset(kwargs.keys())
        return MongoClient(*args, **kwargs)

    return mocker.patch("ampel.core.AmpelDB.MongoClient", side_effect=checky)


@pytest.fixture
def ampel_cli_opts(testing_config, vault):
    return ["--secrets", str(vault), "--config", str(testing_config)]


def test_auth(ampel_cli_opts, check_mongo_auth):

    cli_op, sub_op = T2Command(), "reset"
    parser = cli_op.get_parser(sub_op)
    args, unknown_args = parser.parse_known_args(ampel_cli_opts)

    with pytest.raises(UnknownDatabase):
        cli_op.run(vars(args), unknown_args, sub_op)
    assert check_mongo_auth.called
    for call_args in check_mongo_auth.call_args_list:
        assert {"username", "password"}.issubset(call_args.kwargs.keys())


def test_multi_integer_flags(ampel_cli_opts, check_mongo_auth, mocker: MockerFixture):
    """
    short-option argument parsing handles negative integers
    """

    mock_run = mocker.patch("ampel.cli.T2Command.T2Command.run")

    run("ampel t2 reset -code -5 -7 -2006".split())

    assert mock_run.called
    args, unknown_args, sub_op = mock_run.call_args[0]
    assert not unknown_args, "no unhandled arguments"
    assert args["code"] == [-5, -7, -2006], "all codes assigned to correct option"



"""
This package wraps the [`os.pipe`](https://docs.python.org/3/library/os.html#os.pipe)
simplex communication pipe so it can be used as part of the non-blocking
[`asyncio`](https://docs.python.org/3/library/asyncio.html) event loop. A duplex pipe
is also provided, which allows reading and writing on both ends.

## Simplex example

The following example opens a pipe with the write end in the child process and the read
end in the parent process.

```python3
>>> from multiprocessing import Process
>>> import asyncio
>>>
>>> from aiopipe import aiopipe
>>>
>>> async def main():
...     rx, tx = aiopipe()
...
...     with tx.detach() as tx:
...         proc = Process(target=childproc, args=(tx,))
...         proc.start()
...
...     # The write end is now available in the child process
...     # and detached from the parent process.
...
...     async with rx.open() as rx:
...         msg = await rx.readline()
...
...     proc.join()
...     return msg
>>>
>>> def childproc(tx):
...     asyncio.run(childtask(tx))
>>>
>>> async def childtask(tx):
...     async with tx.open() as tx:
...         tx.write(b"hi from the child process\\n")
>>>
>>> asyncio.run(main())
b'hi from the child process\\n'
>>>
```

## Duplex example

The following example shows a parent and child process sharing a duplex pipe to exchange
messages.

```python3
>>> from multiprocessing import Process
>>> import asyncio
>>>
>>> from aiopipe import aioduplex
>>>
>>> async def main():
...     mainpipe, chpipe = aioduplex()
...
...     with chpipe.detach() as chpipe:
...         proc = Process(target=childproc, args=(chpipe,))
...         proc.start()
...
...     # The second pipe is now available in the child process
...     # and detached from the parent process.
...
...     async with mainpipe.open() as (rx, tx):
...         req = await rx.read(5)
...         tx.write(req + b" world\\n")
...         msg = await rx.readline()
...
...     proc.join()
...     return msg
>>>
>>> def childproc(pipe):
...     asyncio.run(childtask(pipe))
>>>
>>> async def childtask(pipe):
...     async with pipe.open() as (rx, tx):
...         tx.write(b"hello")
...         rep = await rx.readline()
...         tx.write(rep.upper())
>>>
>>> asyncio.run(main())
b'HELLO WORLD\\n'
>>>
```
"""

import os
from asyncio import (
    BaseTransport,
    StreamReader,
    StreamReaderProtocol,
    StreamWriter,
    get_running_loop,
    sleep,
)
from collections.abc import AsyncIterator, Iterator
from contextlib import asynccontextmanager, contextmanager, suppress
from typing import Any

__pdoc__ = {} # type: dict[str,None|str]

def aiopipe() -> tuple["AioPipeReader", "AioPipeWriter"]:
    """
    Create a new simplex multiprocess communication pipe.

    Return the read end and write end, respectively.
    """

    rx, tx = os.pipe()
    return AioPipeReader(rx), AioPipeWriter(tx)

def aioduplex() -> tuple["AioDuplex", "AioDuplex"]:
    """
    Create a new duplex multiprocess communication pipe.

    Both returned pipes can write to and read from the other.
    """

    rxa, txa = aiopipe()
    rxb, txb = aiopipe()

    return AioDuplex(rxa, txb), AioDuplex(rxb, txa)

class AioPipeStream:
    """
    Abstract class for pipe readers and writers.
    """

    __pdoc__["AioPipeStream.__init__"] = None
    __pdoc__["AioPipeStream.open"] = None

    def __init__(self, fd):
        self._fd = fd

    @asynccontextmanager
    async def open(self):
        if self._fd is None:
            raise ValueError("File handle already closed")

        transport, stream = await self._open()

        try:
            yield stream
        finally:
            try:
                transport.close()
                self._fd = None
            except OSError:
                # The transport/protocol sometimes closes the fd before this is reached.
                pass

            # Allow event loop callbacks to run and handle closed transport.
            await sleep(0)

    async def _open(self) -> tuple[BaseTransport, Any]:
        raise NotImplementedError()

    @contextmanager
    def detach(self) -> Iterator["AioPipeStream"]:
        """
        Detach this end of the pipe from the current process in preparation for use in a
        child process.

        This returns a context manager, which must be used as part of a `with` context.
        When the context is entered, the stream is prepared for
        [inheritance](https://docs.python.org/3/library/os.html#fd-inheritance) by the
        child process and returned as the context variable. When the context is exited,
        the stream is closed in the parent process.
        """

        if self._fd is None:
            raise ValueError("File handle already closed")
        try:
            os.set_inheritable(self._fd, True)
            yield self
        finally:
            # NB: close and explicitly invalidate the file handle here. Any call
            # to os.pipe() after this point may create a new file handle with
            # the same value, which we will then happily close whenever the
            # interpreter happens to call __del__()
            self.close()

    def close(self):
        if self._fd is not None:
            os.close(self._fd)
            self._fd = None

    def __del__(self):
        with suppress(OSError):
            self.close()

class AioPipeReader(AioPipeStream):
    """
    The read end of a pipe.
    """

    __pdoc__["AioPipeReader.__init__"] = None
    __pdoc__["AioPipeReader.open"] = """
        Open the receive end on the current event loop.

        This returns an async context manager, which must be used as part of an `async
        with` context. When the context is entered, the receive end is opened and an
        instance of [`StreamReader`][stdlib] is returned as the context variable. When the
        context is exited, the receive end is closed.

        [stdlib]: https://docs.python.org/3/library/asyncio-stream.html#asyncio.StreamReader
    """

    async def _open(self):
        rx = StreamReader()
        transport, _ = await get_running_loop().connect_read_pipe(
            lambda: StreamReaderProtocol(rx),
            os.fdopen(self._fd))

        return transport, rx

class AioPipeWriter(AioPipeStream):
    """
    The write end of a pipe.
    """

    __pdoc__["AioPipeWriter.__init__"] = None
    __pdoc__["AioPipeWriter.open"] = """
        Open the transmit end on the current event loop.

        This returns an async context manager, which must be used as part of an `async
        with` context. When the context is entered, the transmit end is opened and an
        instance of [`StreamWriter`][stdlib] is returned as the context variable. When the
        context is exited, the transmit end is closed.

        [stdlib]: https://docs.python.org/3/library/asyncio-stream.html#asyncio.StreamWriter
    """

    async def _open(self):
        rx = StreamReader()
        transport, proto = await get_running_loop().connect_write_pipe(
            lambda: StreamReaderProtocol(rx),
            os.fdopen(self._fd, "w"))
        tx = StreamWriter(transport, proto, rx, get_running_loop())

        return transport, tx

class AioDuplex:
    """
    Represents one end of a duplex pipe.
    """

    __pdoc__["AioDuplex.__init__"] = None

    def __init__(self, rx: AioPipeReader, tx: AioPipeWriter):
        self._rx = rx
        self._tx = tx

    @contextmanager
    def detach(self) -> Iterator["AioDuplex"]:
        """
        Detach this end of the duplex pipe from the current process in preparation for use
        in a child process.

        This returns a context manager, which must be used as part of a `with` context.
        When the context is entered, the pipe is prepared for
        [inheritance](https://docs.python.org/3/library/os.html#fd-inheritance) by the
        child process and returned as the context variable. When the context is exited,
        the stream is closed in the parent process.
        """

        with self._rx.detach(), self._tx.detach():
            yield self

    @asynccontextmanager
    async def open(self) -> AsyncIterator[tuple["StreamReader", "StreamWriter"]]:
        """
        Open this end of the duplex pipe on the current event loop.

        This returns an async context manager, which must be used as part of an `async
        with` context. When the context is entered, the pipe is opened and the underlying
        [`StreamReader`][reader] and [`StreamWriter`][writer] are returned as the context
        variable. When the context is exited, the pipe is closed.

        [reader]: https://docs.python.org/3/library/asyncio-stream.html#asyncio.StreamReader
        [writer]: https://docs.python.org/3/library/asyncio-stream.html#asyncio.StreamWriter
        """

        async with self._rx.open() as rx, self._tx.open() as tx:
            yield rx, tx



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/dev/DevAmpelContext.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                10.06.2020
# Last Modified Date:  01.04.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Sequence
from typing import Any

from ampel.base.AmpelUnit import AmpelUnit
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.base.LogicalUnit import LogicalUnit
from ampel.config.AmpelConfig import AmpelConfig
from ampel.core.AmpelContext import AmpelContext
from ampel.core.AmpelDB import AmpelDB
from ampel.core.ContextUnit import ContextUnit
from ampel.core.UnitLoader import UnitLoader
from ampel.model.ChannelModel import ChannelModel
from ampel.model.UnitModel import UnitModel
from ampel.util.freeze import recursive_unfreeze
from ampel.util.mappings import set_by_path


class DevAmpelContext(AmpelContext):

	def __init__(self,
		db_prefix: None | str = None, purge_db: bool = False,
		custom_conf: None | dict[str, Any] = None, **kwargs
	) -> None:
		"""
		:db_prefix: customizes the db prefix name
		(ex: "AmpelTest" will result in the databases: AmpelTest_data, AmpelTest_var, AmpelTest_ext)
		:purge_db: wheter all databases with the given prefix should be deleted (purged)
		:extra_conf: convenience parameter which allows to overwrite given parameters of the underlying
		ampel config (possibly frozen) dictionnary. Nested dict keys such as 'general.something' are supported.
		"""

		super().__init__(**kwargs)

		if db_prefix:
			dict.__setitem__(self.config._config['mongo'], 'prefix', db_prefix)  # noqa: SLF001

		if custom_conf or db_prefix:
			conf = self._get_unprotected_conf()
			for k, v in (custom_conf or {}).items():
				set_by_path(conf, k, v)
			self._set_new_conf(conf)

		if purge_db:
			self.db.drop_all_databases()
			self.db.init_db()

		for stored_conf in self.db.get_collection("conf").find({}):
			confid = stored_conf.pop("_id")
			dict.__setitem__(self.config._config["confid"], confid, stored_conf)  # noqa: SLF001


	def add_channel(self, name: int | str, access: Sequence[str] = ()):
		cm = ChannelModel(channel=name, access=access, version=0)
		conf = self._get_unprotected_conf()
		for k, v in cm.__dict__.items():
			set_by_path(conf, f"channel.{name}.{k}", v)
		self._set_new_conf(conf)


	def register_units(self, *Classes: type[AmpelUnit]) -> None:
		for Class in Classes:
			self.register_unit(Class)


	def register_unit(self, Class: type[AmpelUnit]) -> type[AmpelUnit]:

		dict.__setitem__(
			self.config._config['unit'],  # noqa: SLF001
			Class.__name__,
			{
				'fqn': Class.__module__,
				'base': [el.__name__ for el in Class.__mro__[:-1] if 'ampel' in el.__module__],
				'distrib': 'unspecified',
				'file': 'unspecified',
				'version': 'unspecified'
			}
		)

		if issubclass(Class, LogicalUnit | ContextUnit):

			if self.loader._dyn_register is None:  # noqa: SLF001
				self.loader._dyn_register = {}  # noqa: SLF001

			self.loader._dyn_register[Class.__name__] = Class  # noqa: SLF001

		else:
			AuxUnitRegister._dyn[Class.__name__] = Class  # noqa: SLF001

		return Class

	def _get_unprotected_conf(self) -> dict[str, Any]:
		if self.config.is_frozen():
			return recursive_unfreeze(self.config._config) # type: ignore[arg-type]  # noqa: SLF001
		return self.config._config  # noqa: SLF001


	def _set_new_conf(self, conf: dict[str, Any]) -> None:
		self.config = AmpelConfig(conf, True)
		self.db = AmpelDB.new(self.config, self.loader.vault, self.db.require_exists, self.db.one_db)
		self.loader = UnitLoader(self.config, db=self.db, vault=self.loader.vault)


	def new_context_unit(self, unit: str, **kwargs) -> ContextUnit:
		"""
		Convenience method (for notebooks mainly).
		Limitation: config of underlying unit cannot contain/define field 'unit'
		"""
		return self.loader.new_context_unit(
			context = self,
			model = UnitModel(unit = unit, config = kwargs)
		)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/builder/get_env.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                Undefined
# Last Modified Date:  23.04.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import importlib
import sys

from pip._internal.operations import freeze

# ruff: noqa: T201

if __name__ == "__main__":

	exceptions = {

		# Irrevelant
		"typing-extensions", "typing_extensions",
		"prometheus-client", "certifi", "six",

		# Global (added for each unit)
		"pydantic", "pymongo", "ujson", "xxhash"
	}

	pip_env = {
		z[0].replace("-", "_"): z[1]
		for el in freeze.freeze()
		if len(z := el.split("==")) > 1
		and z[0] not in exceptions
	}

	start = set(sys.modules.keys())
	importlib.import_module(sys.argv[1])
	diff = set(sys.modules.keys()) - start

	print(
		{
			m: pip_env[m]
			for k in sorted(diff)
			if (m := k.split(".")[0]) in pip_env
		}
	)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/update/MongoStockIngester.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.12.2017
# Last Modified Date:  21.11.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any

from pymongo import UpdateOne

from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.content.StockDocument import StockDocument
from ampel.mongo.update.HasUpdatesBuffer import HasUpdatesBuffer
from ampel.mongo.utils import maybe_use_each
from ampel.util.collections import try_reduce


class MongoStockIngester(AbsDocIngester[StockDocument], HasUpdatesBuffer):

	def ingest(self, doc: StockDocument) -> None:

		now = doc['journal'][-1]['ts']
		set_on_insert: dict[str, Any] = {'ts.any.tied': now}
		add_to_set = {'channel': maybe_use_each(doc['channel'])}

		if 'name' in doc:
			set_on_insert['name'] = doc['name']
			
		if 'tag' in doc:
			add_to_set['tag'] = maybe_use_each(doc['tag'])

		for el in doc['journal']:
			el['channel'] = try_reduce(el['channel'])

		upd = {
			'$addToSet': add_to_set,
			'$min': {f'ts.{chan}.tied': now for chan in doc['channel']},
			'$max': {f'ts.{chan}.upd': now for chan in doc['channel']},
			'$push': {'journal': maybe_use_each(doc['journal'])}
		}

		if set_on_insert:
			upd['$setOnInsert'] = set_on_insert

		# It should be decided later whether we want to define StockDocument.body
		# as an array and use $push here instead of the (potentially dangerous because
		# associated with data erasure) $set operation
		if 'body' in doc:
			upd['$set'] = {'body': doc['body']}

		# Insert/Update stock document into stock collection
		self.updates_buffer.add_stock_update(
			UpdateOne(
				{'stock': doc['stock'], 'origin': doc['origin']}
				if 'origin' in doc else {'stock': doc['stock']},
				upd,
				upsert=True
			)
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/stage/T3BaseStager.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                08.12.2021
# Last Modified Date:  03.04.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Generator
from time import time
from typing import Any

from ampel.abstract.AbsT3Stager import AbsT3Stager
from ampel.abstract.AbsT3Unit import AbsT3Unit, T
from ampel.base.AmpelUnit import AmpelUnit
from ampel.content.T3Document import T3Document
from ampel.core.DocBuilder import DocBuilder
from ampel.core.EventHandler import EventHandler
from ampel.enum.JournalActionCode import JournalActionCode
from ampel.log.AmpelLogger import AmpelLogger
from ampel.model.UnitModel import UnitModel
from ampel.mongo.update.MongoStockUpdater import MongoStockUpdater
from ampel.struct.T3Store import T3Store
from ampel.struct.UnitResult import UnitResult
from ampel.t3.stage.BaseViewGenerator import BaseViewGenerator
from ampel.types import ChannelId, OneOrMany, StockId, Tag, Traceless, UBson
from ampel.util.mappings import dictify


class T3BaseStager(AbsT3Stager, DocBuilder, abstract=True):
	"""
	Base (abstract) class for several stagers provided by ampel.
	This class does not implement the method stage(...) required by AbsT3Stager,
	it is up to the subclass to do it according to requirements.
	"""


	logger: Traceless[AmpelLogger]
	event_hdlr: Traceless[EventHandler]

	#: tag: tag(s) to add to the :class:`~ampel.content.JournalRecord.JournalRecord` of each selected stock
	extra_journal_tag: None | OneOrMany[Tag] = None

	#: Record the invocation of this event in the stock journal
	update_journal: bool = True

	#: Whether t3 result should be added to t3 store once available
	propagate: bool = True

	#: Applies only for underlying T3ReviewUnits.
	#: Note that if True, a T3 document will be created even if a t3 unit returns None
	save_stock_ids: bool = False


	def __init__(self, **kwargs) -> None:

		super().__init__(**kwargs)
		self.stock_updr = MongoStockUpdater(
			ampel_db = self.context.db,
			tier = 3,
			run_id = self.event_hdlr.get_run_id(),
			process_name = self.event_hdlr.process_name,
			logger = self.logger,
			raise_exc = self.event_hdlr.raise_exc,
			extra_tag = self.extra_journal_tag,
			update_journal = self.update_journal,
			bump_updated = False
		)


	def get_unit(self, unit_model: UnitModel, chan: None | OneOrMany[ChannelId] = None) -> AbsT3Unit:

		# new_safe_logical_unit returns a T3 unit instantiated with
		# a logger based on DefaultRecordBufferingHandler
		return self.context.loader.new_safe_logical_unit(
			unit_model,
			unit_type = AbsT3Unit,
			logger = self.logger,
			_chan = self.channel or chan # type: ignore[arg-type] # to be improved when time allows
		)


	def handle_t3_result(self,
		t3_unit: AbsT3Unit,
		res: UBson | UnitResult,
		t3s: T3Store,
		stocks: None | list[StockId],
		ts: float,
		log_extra: None | dict[str, Any] = None
	) -> None | T3Document:

		# Let's consider logs as a result product
		if (buf_hdlr := getattr(t3_unit, '_buf_hdlr', None)) and buf_hdlr.buffer:
			buf_hdlr.forward(self.logger, extra=log_extra)

		if isinstance(res, UnitResult):
			if stocks and res.journal:
				self.stock_updr.add_journal_record(
					stock = stocks, # used to match stock docs
					jattrs = res.journal,
					unit = t3_unit.__class__.__name__,
					action_code = JournalActionCode.T3_ADD_DOC
				)
			if res.body is not None or res.code is not None:
				return self.craft_t3_doc(t3_unit, res, t3s, ts, stocks)
		elif res is not None or (res is None and self.save_stock_ids and stocks):
			return self.craft_t3_doc(t3_unit, res, t3s, ts, stocks)

		return None


	def craft_t3_doc(self,
		t3_unit: AmpelUnit,
		res: None | UBson | UnitResult,
		t3s: T3Store,
		ts: float,
		stocks: None | list[StockId] = None
	) -> T3Document:

		t3d = super().craft_doc(self.event_hdlr, t3_unit, res, ts, doc_type=T3Document)
		if self.save_stock_ids and stocks:
			t3d['stock'] = stocks

		if t3s.session:
			t3d['session'] = dictify(t3s.session)

		return t3d


	def proceed(self,
		t3_unit: AbsT3Unit,
		view_generator: BaseViewGenerator[T],
		t3s: T3Store
	) -> Generator[T3Document, None, None]:
		"""
		Executes method 'process' of t3 unit with provided views generator and t3 store.
		Handles potential t3 unit result as well.
		"""

		try:
			self.logger.info("Running T3 unit", extra={'unit': t3_unit.__class__.__name__})
			ts = time()
			if (
				((ret := t3_unit.process(view_generator, t3s)) or self.save_stock_ids) and
				(x := self.handle_t3_result(t3_unit, ret, t3s, view_generator.get_stock_ids(), ts))
			):
				yield x
		except Exception as e:
			self.event_hdlr.handle_error(e, self.logger)
		finally:
			if self.stock_updr.update_journal:
				self.stock_updr.flush()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/LogCommand.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                15.03.2021
# Last Modified Date:  27.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from argparse import ArgumentParser
from collections.abc import Sequence
from typing import Any

from ampel.abstract.AbsIdMapper import AbsIdMapper
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser
from ampel.cli.ArgParserBuilder import ArgParserBuilder
from ampel.cli.LoadJSONAction import LoadJSONAction
from ampel.cli.MaybeIntAction import MaybeIntAction
from ampel.core.AmpelContext import AmpelContext
from ampel.log.LogsDumper import LogsDumper
from ampel.mongo.query.var.LogsLoader import LogsLoader
from ampel.mongo.query.var.LogsMatcher import LogsMatcher

hlp = {
	'show': 'Print logs to stdout',
	'save': 'Save logs',
	'tail': 'Like tail -f. Query interval can be customized with -refresh-rate (default: 1.0 second)\n',
	'config': 'Path to an ampel config file (yaml/json)',
	'secrets': 'Path to a YAML secrets store in sops format',
	'db': 'Database prefix. If set, "-mongo.prefix" value will be ignored',
	'after': 'Match logs emitted after the provided date-time string [%%]',
	'before': 'Match logs emitted before the provided date-time string [%%]',
	'stock': 'Stock id(s) [o]',
	'channel': 'Channel(s) [o]',
	'run': 'Integer run id(s) [o]',
	'run-json': 'Matching dict for run id(s) |§|',
	'custom-key': 'Custom key to match (ex: "a" for alert id(s)). Parameter -custom-value must be set as well',
	'custom-value': 'Value(s) associated with -custom-key to be matched against [o]',
	'flag': 'Integer flag value. The $bitsAllSet matching operator is used.\nCan be used to filter based on "tier"',
	'date-format': 'Use custom date-time format for the logs output (ex: %%Y-%%m-%%d %%H:%%M:%%S)', # Note: % need to be escaped
	'id-mapper': 'Convert stock ids using the provided id mapper (ex: ZTFIdMapper)',
	'no-resolve-flag': 'Keep flag as int rather than performing str conversion\n' +
		'(ex: prints 1169 instead of "INFO|CORE|SCHEDULED_RUN|T0")',
	'no-resolve-stock': 'Keep stock as int when matching using id-mapper',
	'out': 'Path to file were output will be written (printed to stdout otherwise)',
	'to-json': 'Output json structure',
	'to-pretty-json': 'Output json structure with spacier formatting than the default one',
	'main-separator': 'Main separator (space by default, ex: 2021-03-16T22:11:09.000Z INFO...)',
	'extra-separator': 'Extra separator (space by default, ex: [run=1 stock=9])',
	'refresh-rate': 'Tail refresh interval in seconds (float, default: 1.0 second)',
	'verbose': 'Increases verbosity'
}

class LogCommand(AbsCoreCommand):
	"""
	TODO: Register and catch interupts in tail mode to avoid KeyboardInterrupt stack
	"""

	@staticmethod
	def get_sub_ops() -> list[str]:
		return ['show', 'tail', 'save']

	# Mandatory implementation
	def get_parser(self, sub_op: None | str = None) -> ArgumentParser | AmpelArgumentParser:

		if sub_op in self.parsers:
			return self.parsers[sub_op]

		sub_ops = self.get_sub_ops()
		if sub_op is None or sub_op not in sub_ops:
			return AmpelArgumentParser.build_choice_help(
				'log', sub_ops, hlp,
				description = 'Match and view, export or tail AMPEL logs'
			)

		builder = ArgParserBuilder('log')
		builder.add_parsers(sub_ops, hlp)

		builder.notation_add_note_references()
		builder.notation_add_example_references()

		# Required
		builder.opt('config', type=str)
		builder.req('out', 'save', type=str)

		# Optional
		builder.opt('secrets', default=None)
		builder.opt('db', default=None)
		builder.opt('id-mapper', type=str)

		#parser.arg(g, 'tail', action='store', metavar='#', default=None, const=1.0, type=float, nargs='?')
		builder.opt('verbose', action='count', default=0)
		builder.opt('debug', action='count', default=0, help='Debug')
		builder.opt(
			'refresh-rate', 'tail', action='store', metavar='#',
			const=1.0, nargs='?', type=float, default=1.0
		)
		
		# Optional match criteria
		builder.add_group('match', 'Optional matching criteria [&]', sub_ops='all')
		builder.arg('after', group='match', sub_ops='show|save', type=str) # does not work with tail
		builder.arg('before', group='match', sub_ops='show|save', type=str)
		builder.arg('channel', group='match', action=MaybeIntAction, nargs='+')
		builder.arg('stock', group='match', action=MaybeIntAction, nargs='+')

		builder.xargs(
			group='match', sub_ops='show|save', xargs = [
				{'name': 'run', 'action': MaybeIntAction, 'nargs': '+'},
				{'name': 'run-json', 'action': LoadJSONAction, 'metavar': '#', 'dest': 'run'}
			]
		)
		builder.arg('flag', group='match', type=int, nargs=1)
		builder.arg('custom-key', group='match', type=str)
		builder.arg('custom-value', group='match', action=MaybeIntAction, nargs='+')

		builder.add_group('out', 'Optional output parameters', sub_ops='all')
		builder.xargs(
			group='out', sub_ops='all', xargs = [
				{'name': 'to-json', 'action': 'store_true'},
				{'name': 'to-pretty-json', 'action': 'store_true'}
			]
		)

		builder.add_group('format', 'Optional global format parameters', sub_ops='all')
		builder.arg('date-format', group='format', type=str)
		builder.arg('no-resolve-flag', group='format', dest='resolve_flag', action='store_false')
		builder.set_group_defaults('format', sub_ops='all', resolve_flag=True)
		builder.arg('no-resolve-stock', group='format', action='store_true')

		builder.add_group('format2', 'Optional specific format parameters', sub_ops='all')
		builder.arg(
			'main-separator', group='format2', sub_ops='all',
			type=str, default=' ', const=' ', nargs='?'
		)
		builder.arg(
			'extra-separator', group='format2', sub_ops='all',
			type=str, default=' ', const=' ', nargs='?'
		)

		# Notes
		builder.hint_all_query_logic(ref='&')
		builder.hint_time_format('show|save', ref='%')
		builder.note('multi-values are OR-matched', ref='o')
		builder.hint_all_config_override()

		# Examples
		for el in sub_ops:
			p = f'ampel log {el} '
			a = ' -out /path/to/file.txt' if el == 'save' else ''
			builder.example(el, "-run-json '{\"$gt\": 12}'", ref='§', prepend=p, append=a)
			builder.example(el, '-stock 85628462', prepend=p, append=a)
			builder.example(el, '-db MyDB -to-json', prepend=p, append=a)
			builder.example(el, '-run 8 -db AmpelTest', prepend=p, append=a)
			if el != 'tail':
				builder.example(el, '-after 2020-11-03T12:12:00', prepend=p, append=a)
			builder.example(el, '-stock ZTF17aaatbxz ZTF20aaquaxr -id-mapper ZTFIdMapper', prepend=p, append=a)
			builder.example(el, '-custom-key unit -custom-value T3DemoSavePlot -to-pretty-json -no-resolve-flag', prepend=p, append=a)
			builder.example(el, '-stock ZTF20aaquast -id-mapper ZTFIdMapper', prepend=p, append=a)

		builder.example('tail', '-refresh-rate 10')

		self.parsers.update(
			builder.get()
		)

		return self.parsers[sub_op]


	# Mandatory implementation
	def run(self, args: dict[str, Any], unknown_args: Sequence[str], sub_op: None | str = None) -> None:

		ctx = self.get_context(
			args, unknown_args, ContextClass=AmpelContext,
			require_existing_db = args['db'] or True, one_db='auto'
		)

		self.flag_strings: dict = {}

		if (
			(args['custom_key'] and not args['custom_value']) or
			(args['custom_value'] and not args['custom_key'])
		):
			raise ValueError(
				'Both parameter "--custom-key" and "--custom-value"' +
				' must be used when either one is requested'
			)

		if args['custom_key']:
			args['custom'] = {args['custom_key']: args['custom_value']}

		if args['id_mapper']:
			args['id_mapper'] = AuxUnitRegister.get_aux_class(
				args['id_mapper'], sub_type=AbsIdMapper
			)()

		matcher = LogsMatcher.new(**args)
		loader = LogsLoader(
			**(
				args |
				{'datetime_ouput': 'date' if args['date_format'] else 'string'}
			)
		)

		if args['no_resolve_stock']:
			args['id_mapper'] = None

		ld = LogsDumper(**args)
		col = ctx.db.get_collection('log')
		mcrit = matcher.get_match_criteria()

		if sub_op == 'tail':
			import time
			from datetime import datetime, timezone

			from bson.objectid import ObjectId
			mcrit['_id'] = {'$gte': ObjectId.from_datetime(datetime.now(tz=timezone.utc))}
			ld.datetime_key = 'date'
			loader.datetime_key = 'date'

			while True:
				log_entries = loader.fetch_logs(col, mcrit)
				if log_entries:
					next_match = log_entries[-1]['_id']
					ld.process(log_entries) # type: ignore[arg-type]
					mcrit['_id'] = {'$gt': next_match}
				else:
					time.sleep(args['refresh_rate'])

		print(mcrit)
		log_entries = loader.fetch_logs(col, mcrit)
		ld.process(log_entries) # type: ignore[arg-type]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/util/logicschema.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                03.04.2021
# Last Modified Date:  03.04.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Sequence
from typing import Any, Literal

from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.OneOf import OneOf
from ampel.types import T, strict_iterable
from ampel.util.collections import check_seq_inner_type


def to_logical_dict(v, field_name: int | str | dict[str, Any]) -> dict[str, Any]:
	"""
	Converts str/int into {'any_of': int/str}.
	Checks structure in case dict is provided
	(how deep potential nested logic are, raises error in case of incorrect logic).
	:raises ValueError: in case provided argument is incorrect
	"""

	if isinstance(v, list):
		raise ValueError(
			f"stock->select->{field_name} config error\n" +
			f"'{field_name}' parameter cannot be a list. " +
			"Please use the following syntax:\n" +
			" -> {'any_of': ['Ab', 'Cd']} or\n" +
			" -> {'all_of': ['Ab', 'Cd']} or\n" +
			" -> {'one_of': ['Ab', 'Cd']} or\n" +
			"One level nesting is allowed, please see\n" +
			"conditional_expr_converter(..) docstring for more info"
		)

	if isinstance(v, str | int):
		return {'any_of': [v]}

	if isinstance(v, dict):

		if len(v) != 1:
			raise ValueError(
				f"stock->select->{field_name} config error\n" +
				f"Unsupported dict format {v}"
			)

		if 'any_of' in v:

			if not isinstance(v['any_of'], strict_iterable):
				raise ValueError(
					f"stock->select->{field_name}:any_of config error\n" +
					"Invalid dict value type: {}. Must be a sequence".format(type(v['any_of']))
				)

			# 'any_of' supports only a list of dicts and str/int
			if not check_seq_inner_type(v['any_of'], (str, int, dict), multi_type=True):
				raise ValueError(
					f"stock->select->{field_name}:any_of config error\n" +
					"Unsupported nesting (err 2)"
				)

			if not check_seq_inner_type(v['any_of'], (int, str)) and len(v['any_of']) < 2:
				raise ValueError(
					f"stock->select->{field_name}:any_of config error\n" +
					"any_of list must contain more than one element when containing all_of\n" +
					f"Offending value: {v}"
				)

			for el in v['any_of']:

				if isinstance(el, dict):

					if 'any_of' in el:
						raise ValueError(
							f"stock->select->{field_name}:any_of.any_of config error\n" +
							"Unsupported nesting (any_of in any_of)"
						)

					if 'all_of' in el:

						# 'all_of' closes nesting
						if not check_seq_inner_type(el['all_of'], (int, str)):
							raise ValueError(
								f"stock->select->{field_name}:any_of.all_of config error\n" +
								"Unsupported nesting (all_of list content must be str/int)"
							)

						if len(set(el['all_of'])) < 2:
							raise ValueError(
								f"stock->select->{field_name}:all_of config error\n" +
								"Please do not use all_of with just one element\n" +
								f"Offending value: {el}"
							)

					else:
						raise ValueError(
							f"stock->select->{field_name}:any_of config error\n" +
							f"Unsupported nested dict: {el}"
						)

		elif 'all_of' in v:

			# 'all_of' closes nesting
			if (
				not isinstance(v['all_of'], strict_iterable) or
				not check_seq_inner_type(v['all_of'], (int, str))
			):
				raise ValueError(
					f"stock->select->{field_name}:all_of config error\n"
					f"Invalid type for value {v['all_of']}\n(must be a sequence, is: {type(v['all_of'])})\n"
					"Note: no nesting is allowed below 'all_of'"
				)

			if len(set(v['all_of'])) < 2:
				raise ValueError(
					f"stock->select->{field_name}:all_of config error\n"
					"Please do not use all_of with just one element\n"
					f"Offending value: {v}"
				)

		elif 'one_of' in v:

			# 'one_of' closes nesting
			if (
				not isinstance(v['one_of'], strict_iterable) or
				not check_seq_inner_type(v['one_of'], (int, str))
			):
				raise ValueError(
					f"stock->select->{field_name}:one_of config error\n"
					f"Invalid type for value {v['one_of']}\n(must be a sequence, is: {type(v['one_of'])})\n"
					"Note: no nesting is allowed below 'one_of'"
				)

		else:
			raise ValueError(
				f"stock->select->{field_name} config error\n" +
				"Invalid dict key (only 'any_of', 'all_of', 'one_of' are allowed)"
			)

	return v


def reduce_to_set(
	arg: (
		T |
		# unsure if mypy understands unions of dicts with different key literals actually
		dict[Literal['all_of', 'one_of'], Sequence[T]] |
		dict[Literal['any_of'], Sequence[T] | dict[Literal['all_of'], Sequence[T]]] |
		AllOf[T] | AnyOf[T] | OneOf[T]
	),
	in_type: tuple[type, ...] = (str, int)
) -> set[T]:
	"""
	.. sourcecode:: python\n
		for schema in (a,b,c,d,e):
			print("Schema: %s" % schema)
			print("Reduced set: %s" % reduce_to_set(schema))

		Schema: 'a'
		Reduced set: {'a'}
		Schema: {'any_of': ['a', 'b', 'c']}
		Reduced set: {'b', 'a', 'c'}
		Schema: {'all_of': [1, 2, 3]}
		Reduced set: {1, 2, 3}
		Schema: {'any_of': [{'all_of': ['a', 'b']}, 'c']}
		Reduced set: {'b', 'a', 'c'}
		Schema: {'any_of': [{'all_of': ['a', 'b']}, {'all_of': ['a', 'c']}, 'd']}
		Reduced set: {'d', 'b', 'a', 'c'}
	"""

	if isinstance(arg, in_type):
		return {arg} # type: ignore[arg-type]

	if isinstance(arg, AllOf | AnyOf | OneOf):
		v: Any = arg.dict()

	elif isinstance(arg, dict):
		v = arg

	else:
		raise ValueError(f"Unsupported arg type: {type(arg)}")

	if "any_of" in v:
		s: set[T] = set()
		for el in v['any_of']:
			if isinstance(el, in_type):
				s.add(el) # type: ignore[arg-type]
			elif isinstance(el, dict):
				for ell in next(iter(el.values())):
					s.add(ell)
			else:
				raise ValueError("unsupported format (1)")
		return s

	if 'all_of' in v:
		return set(v['all_of'])

	if 'one_of' in v:
		return set(v['one_of'])

	raise ValueError("unsupported format (2)")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/include/session/T3SessionAlertsNumber.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                06.01.2020
# Last Modified Date:  17.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import ClassVar

from ampel.abstract.AbsT3Supplier import AbsT3Supplier
from ampel.mongo.query.var.events import build_t0_stats_query
from ampel.struct.T3Store import T3Store
from ampel.t3.include.session.T3SessionLastRunTime import T3SessionLastRunTime


class T3SessionAlertsNumber(AbsT3Supplier[dict]):
	""" Note: also returns "last run time" of process """

	key: ClassVar[str] = "processed_alerts"


	def supply(self, t3s: T3Store) -> dict:
		""" Returns number of alerts processed since last run of this event as "processed_alerts" """

		d = T3SessionLastRunTime(
			context = self.context,
			logger = self.logger,
			event_hdlr = self.event_hdlr
		).supply(t3s)

		if not d[T3SessionLastRunTime.key]:
			self.logger.info(
				"Last run time not available, cannot determine " +
				"how many alerts were processed since then"
			)
			d[self.key] = None
			return d

		# Get number of alerts processed since last run
		res = next(
			self.context.db.get_collection('event').aggregate(
				build_t0_stats_query(gte_time=d[T3SessionLastRunTime.key])
			),
			None
		)

		d[self.key] = 0 if res is None else res.get('alerts')
		return d



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/collector/LoggingCollector.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                03.03.2020
# Last Modified Date:  18.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any

from ampel.config.collector.AbsDictConfigCollector import AbsDictConfigCollector
from ampel.config.collector.ConfigCollector import ConfigCollector
from ampel.log import VERBOSE
from ampel.log.handlers.AmpelStreamHandler import AmpelStreamHandler
from ampel.mongo.update.var.DBLoggingHandler import DBLoggingHandler


class LoggingCollector(AbsDictConfigCollector):

	def add(self,
		arg: dict[str, Any],
		dist_name: str,
		version: str | float | int,
		register_file: str
	) -> None:

		# validate model
		if self.verbose:
			self.logger.log(VERBOSE, "Validating logging configuration")

		try:

			# Loop through logging profiles
			for profile, config in arg.items():

				if "console" in config:
					try:
						AmpelStreamHandler(**config['console'])
					except Exception as e:
						self.error(
							f"Incorrect console logging configuration for: {config['console']} " +
							ConfigCollector.distrib_hint(dist_name, register_file) + ": \n" + str(e)
						)
						continue

				if "db" in config:
					try:
						DBLoggingHandler.validate(config['db'])
					except Exception as e:
						self.error(
							f"Incorrect db logging configuration for: {config['db']} " +
							ConfigCollector.distrib_hint(dist_name, register_file) + ": \n" + str(e)
						)
						continue

				self.__setitem__(profile, config)

				if self.verbose:
					self.logger.log(VERBOSE, f"Adding logging profile '{profile}'")

		except Exception as e:
			print(e)
			self.error(
				"Incorrect logging configuration " +
				ConfigCollector.distrib_hint(dist_name, register_file)
			)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/JobCommand.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                15.03.2021
# Last Modified Date:  05.04.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import filecmp
import importlib
import importlib.metadata
import io
import os
import platform
import shutil
import signal
import subprocess
import sys
import tempfile
from argparse import ArgumentParser
from collections.abc import Sequence
from contextlib import suppress
from multiprocessing import Process, Queue
from time import sleep, time
from typing import Any

import psutil
import ujson
import yaml

from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser
from ampel.cli.config import get_user_data_config_path
from ampel.cli.MaybeIntAction import MaybeIntAction
from ampel.cli.utils import _maybe_int
from ampel.core.AmpelContext import AmpelContext
from ampel.core.EventHandler import EventHandler
from ampel.dev.DevAmpelContext import DevAmpelContext
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.LogFlag import LogFlag
from ampel.model.job.JobModel import JobModel
from ampel.model.job.JobTaskModel import JobTaskModel
from ampel.model.UnitModel import UnitModel
from ampel.util.collections import try_reduce
from ampel.util.debug import MockPool
from ampel.util.distrib import get_dist_names
from ampel.util.freeze import recursive_freeze
from ampel.util.getch import yes_no
from ampel.util.hash import build_unsafe_dict_id
from ampel.util.pretty import get_time_delta, out_stack, prettyjson
from ampel.util.template import apply_templates


class JobCommand(AbsCoreCommand):
	"""
	Processes ampel "jobs" (yaml files).
	A job:
	- is a sharable analysis shema
	- contains configurations of processor units to be run sequentially
	- should also contain everything required to run the underlying processors, that is:
	  channels or aliases definitions and custom resources potentially required by processors.
	- requires access to an ampel config file since the database, unit and resource definitions
	  defined therein are necessary (might be improved in the future)
	"""

	def __init__(self) -> None:
		self.parser = None
		super().__init__()

	@staticmethod
	def get_sub_ops() -> None | list[str]:
		return None

	def get_cli_op_name(self) -> str:
		return "job"

	# Mandatory implementation
	def get_parser(self, sub_op: None | str = None) -> ArgumentParser | AmpelArgumentParser:

		if self.parser:
			return self.parser

		op = self.get_cli_op_name()
		parser = AmpelArgumentParser(op)
		parser.args_not_required = True
		parser.set_help_descr({
			'debug': 'Debug',
			#'verbose': 'increases verbosity',
			'config': 'path to an ampel config file (yaml/json)',
			'schema': 'path to YAML job file (multiple files will be aggregated)',
			'secrets': 'path to a YAML secrets store in sops format',
			'keep-db': 'do not reset databases even if so requested by job file',
			'reset-db': 'reset databases even if not requested by job file',
			'no-conf-check': 'do not check for central config changes',
			'no-agg': 'enables display of matplotlib plots via plt.show() for debugging',
			'no-breakpoint': 'ignore pdb breakpoints',
			'no-mp': 'deactivates multiprocessing by monkey patching multiprocessing.Pool with ampel.util.debug.MockPool for debugging',
			'interactive': 'you will be asked for each task whether it should be run or skipped\n' + \
				'and - if applicable - if the db should be reset. Option -task will be ignored.',
			'edit': 'edit job schema(s) before execution. Customize env variable EDITOR if need be. Default: "raw":\n' +
					'"raw": edit raw schema file(s) before parsing\n' +
					'"parsed": edit (possibly aggregated) schema after parsing by the yaml interpreter\n' +
					'"model": edit job model after templating (if applicable). Only task-related changes will be taken into account',
			'keep-edits': 'do not remove edited job schemas from temp dir',
			'fzf': 'choose schema file using fzf (linux/max only, fzf command line utility must be installed)',
			'task': 'only execute task(s) with provided index(es) [starting with 0].\n' +
					'Value "last" is supported. Ignored if -interactive is used',
			'show-plots': 'show plots created by job (requires ampel-plot-cli. Use "export AMPEL_PLOT_DPI=300" to increase png quality)',
			#'allow-resource-override': 'allow t3 units to overwrite resources previously set by other t3 units',
			'show-plots-cmd': 'show command required to show plots created by job (requires ampel-plot-cli)',
			'wait-pid': 'wait until process with PID completes before processing current job',
			'print-schema': 'print (potentially edited) schema before execution',
			'print-schema-after': 'print (potentially edited) schema after execution',
			'stdin': 'read job schema from stdin',
		})

		parser.req('config', type=str)
		parser.opt('schema', nargs='+')
		parser.opt('task', action=MaybeIntAction, nargs='+')
		parser.opt('interactive', action='store_true')
		parser.opt('debug', action='store_true')
		parser.opt('edit', nargs='?', default=None, const='raw')
		parser.opt('keep-edits', action='store_true')
		parser.opt('fzf', action='store_true')
		parser.opt('keep-db', action='store_true')
		parser.opt('reset-db', action='store_true')
		parser.opt('max-parallel-tasks', type=int, default=os.cpu_count())
		parser.opt('no-conf-check', action='store_true')
		parser.opt('no-agg', action='store_true')
		parser.opt('no-breakpoint', action='store_true')
		parser.opt('no-mp', action='store_true')
		#parser.opt('allow-resource-override', action='store_true')
		parser.opt('show-plots', action='store_true')
		parser.opt('show-plots-cmd', action='store_true')
		parser.opt('secrets', type=str)
		parser.opt('wait-pid', type=int, default=0)
		parser.opt('print-schema', action='store_true')
		parser.opt('print-schema-after', action='store_true')
		parser.opt('stdin', action='store_true')

		# Example
		parser.example(f'{op} job_file.yaml')
		parser.example(f'{op} job_part1.yaml job_part2.yaml')
		parser.example(f'{op} -keep-db -task last job_file.yaml')
		parser.example(f'{op} -show-plots job.yaml')
		parser.example(f'{op} -fzf  [requires fzf command line utility]')
		parser.example(f'pbpaste | ampel {op} -stdin -no-conf-check -show-plots', prepend="")
		return parser


	# Mandatory implementation
	def run(self, args: dict[str, Any], unknown_args: Sequence[str], sub_op: None | str = None) -> None:

		start_time = time()
		psys = platform.system().lower()
		logger = AmpelLogger.get_logger(base_flag=LogFlag.MANUAL_RUN)

		if args['no_mp']:
			"""
			# ThreadPool (dummy.Pool) is not enough in some circunstances
			# as matplotlib has issues when run outside main thread
			def Pool(**kwargs):
				kwargs['processes'] = 1
				return multiprocessing.dummy.Pool(**kwargs)
			"""
			import multiprocessing
			multiprocessing.__dict__['Pool'] = MockPool
			if (
				not args['no_agg'] and
				yes_no('Set -no-agg option too (required for matplotlib interactions)')
			):
				args['no_agg'] = True

		if not args['no_agg']:
			try:
				import matplotlib as mpl
				mpl.use('Agg')
			except Exception:
				pass

		if args['no_breakpoint']:
			os.environ["PYTHONBREAKPOINT"] = "0"

		if not isinstance(args['task'], list) and args['task'] is not None:
			args['task'] = [args['task']]

		schema_paths = args['schema'] or [
			unknown_args.pop(unknown_args.index(el)) # type: ignore[attr-defined]
			for el in list(unknown_args)
			# This might conflict with very special - yet never used - config overrides
			if el.endswith('yml') or el.endswith('yaml')
		]

		if args.get('fzf') and not schema_paths and psys.lower() != "windows":

			if shutil.which('fzf') is None:
				with out_stack():
					raise ValueError("Option '-fzf' requires the fzf command line utility")

			p1 = subprocess.Popen(
				['find', '.', '-name', '*.yml', '-o', '-name', '*.yaml'],
				stdout = subprocess.PIPE
			)
			p2 = subprocess.Popen('fzf', stdin=p1.stdout, stdout=subprocess.PIPE)
			out, err = p2.communicate()

			if err:
				with out_stack():
					raise ValueError("Error occured during fzf execution")

			if (selection := out.decode('utf8').strip()):
				schema_paths = [selection]
			else:
				print("No schema selected")
				return

		if args.get('edit'):
			tmp_files: list[str] = []

		if args.get('edit') == 'raw':

			if schema_paths:
				for sfile in schema_paths:
					_, fname = tempfile.mkstemp(suffix='.yml', text=True)
					shutil.copyfile(sfile, fname)
					edit_job(fname)
					tmp_files.append(fname)

					if filecmp.cmp(sfile, fname):
						logger.info(f'Using original schema: {sfile}')
						os.unlink(fname)
						continue

					schema_paths[schema_paths.index(sfile)] = fname
					logger.info(f'Using modified schema: {fname}')
			else:
				schema_paths = [
					edit_job(tempfile.mkstemp(suffix='.yml')[1])
				]

		if not schema_paths and not args.get('stdin'):
			self.get_parser().print_help()
			return

		if args.get('stdin'):
			job, _ = self.get_job_schema(schema_content=sys.stdin.read(), compute_sig=False)
		else:
			job, _ = self.get_job_schema(schema_paths, compute_sig=False)

		if job is None:
			return

		if args.get('edit') == 'parsed':
			fd, fname = tempfile.mkstemp(suffix='.yml')
			# Seems fd does not work with yaml.dump(), unsure why
			with open(fname, "w") as f:
				yaml.dump(
					ujson.loads(job.json(exclude_unset=True)), f,
					sort_keys=False, default_flow_style=None
				)
			edit_job(fname)
			tmp_files.append(fname)
			job, _ = self.get_job_schema(schema_paths=[fname], compute_sig=False)

		if job is None:
			return

		schema_descr = '|'.join(
			[os.path.basename(sf) for sf in schema_paths]
		).replace('.yaml', '').replace('.yml', '')

		if len(schema_paths) > 1:
			logger.info(f'Running job using composed schema: {schema_descr}')

		# Check or set env variable(s)
		for psys in ('any', platform.system().lower()):
			if psys in job.env:
				for k, v in job.env[psys].check.items():
					if k not in os.environ:
						logger.info(f'Environment variable {k}={v} required')
						return
					if v and v != _maybe_int(os.environ[k]):
						logger.info(
							f'Environment variable {k} value mismatch:\n'
							f' required: {v} ({type(v)})\n'
							f' set: {os.environ[k]}'
						)
						return
				for k, v in job.env[psys].set.items():
					logger.info(f'Setting local environment variable {k}={v}')
					os.environ[k] = str(v)

		purge_db = job.mongo.reset or args['reset_db']
		if purge_db and args['keep_db']:
			logger.info('Keeping existing databases as option "-keep-db" was specified')
			purge_db = False

		if args['interactive']:

			try:
				if purge_db:
					purge_db = yes_no('Delete existing databases')

				args['task'] = []
				for i, td in enumerate(job.task):
					s = f' [{td.title}]' if td.title else ''
					if yes_no(f'Process task #{i}' + s):
						args['task'].append(i)
			except KeyboardInterrupt:
				sys.exit()

		if args['task']:

			if 'last' in args['task']:
				args['task'].remove('last')
				args['task'].append(len(job.task) - 1)

			if sum(args['task']) > 0 and not args['interactive'] and purge_db and not args['reset_db']:
				logger.info('Ampel job file requires db reset but argument "task" was provided')
				logger.info('Please add argument -reset-db to confirm you are absolutely sure...')
				return

		if job.requirements:
			dist_names = [el.lower() for el in get_dist_names()]
			missing = [dist for dist in job.requirements if dist.lower() not in dist_names]
			if missing:
				logger.info(f'Please install {try_reduce(missing)} to run this job\n')
				return # raise ValueError ?

		if (
			not args['config'] and
			not os.path.exists(get_user_data_config_path()) and
			yes_no('Config seems to be missing, build and install')
		):
			from ampel.cli.ConfigCommand import ConfigCommand
			cc = ConfigCommand()
			a, ua = cc.get_parser('install').parse_known_args()
			cc.run(vars(a), ua, sub_op = 'install')

		s = f'Running job {job.name or schema_descr}'
		logger.info(s, extra={'pid': os.getpid()})

		print(' ' + '-'*len(s))

		if args['print_schema']:
			print('\nJob schema:\n')
			print('#'*50)
			print(
				'\n' + yaml.dump(
					job.dict(exclude_unset=True),
					sort_keys=False, default_flow_style=None
				)
			)
			print('#'*50)

		ctx = self.get_context(
			args, unknown_args, logger,
			freeze_config = False,
			ContextClass = DevAmpelContext,
			purge_db = purge_db,
			db_prefix = job.mongo.prefix,
			require_existing_db = False,
			one_db = True
		)

		config_dict = ctx.config._config  # noqa: SLF001

		# Check for outdated config
		if 'build' in config_dict and not args['no_conf_check']:
			for k in config_dict['build']:
				if 'ampel-' in k:
					config_v = config_dict['build'][k]
					current_v = importlib.metadata.distribution(k).version
					if (
						config_v != current_v and
						yes_no(
							f'\nVersion of {k} has changed since config was last built '
							f'(config: {config_v}, current: {current_v}), '
							f'rebuild and install new config'
						)
					):
						from ampel.cli.ConfigCommand import ConfigCommand
						cc = ConfigCommand()
						a, ua = cc.get_parser('install').parse_known_args()
						cc.run(vars(a), ua, sub_op = 'install')

						# Reload config
						ctx = self.get_context(
							args, unknown_args, logger,
							freeze_config = False,
							ContextClass = DevAmpelContext,
							purge_db = purge_db,
							db_prefix = job.mongo.prefix,
							require_existing_db = False,
							one_db = True
						)
						break

		self._patch_config(config_dict, job, logger)
		ctx.config._config = recursive_freeze(config_dict)  # noqa: SLF001

		# Ensure that job content saved in DB reflects options set dynamically
		if args['task']:
			for i in range(len(job.task)):
				if i not in args['task']:
					logger.info(f'Skipping task #{i} as requested')
			job.task = [task for i, task in enumerate(job.task) if i in args['task']]

		# Morphes tasks as well (templates)
		jtasks = self.load_tasks(ctx, job, logger, args.get('debug', False))

		# recreate JobModel with templates resolved
		job_dict = self.get_job_dict(job, jtasks)

		if args.get('edit') == 'model':
			fd, fname = tempfile.mkstemp(suffix='.yml')
			# Seems fd does not work with yaml.dump(), unsure why
			with open(fname, "w") as f:
				yaml.dump(job_dict, f, sort_keys=False, default_flow_style=None)
			edit_job(fname)
			tmp_files.append(fname)
			with open(fname) as f:
				job_dict = yaml.safe_load(f)
			jtasks = job_dict['task']

		if args.get('edit') and not args.get('keep_edits'):
			for el in tmp_files:
				with suppress(BaseException):
					os.unlink(el)

		if (wpid := args['wait_pid']) and psutil.pid_exists(wpid):
			logger.info(f'Waiting until process with PID {wpid} completes')
			while (psutil.pid_exists(wpid)):
				sleep(5)
			start_time = time()

		logger.info('Saving job schema')
		job_sig = build_unsafe_dict_id(job_dict, size=-64)
		job.sig = job_sig
		ctx.db.get_collection('job').update_one(
			{'_id': job_sig},
			{'$setOnInsert': job_dict},
			upsert=True
		)

		# Heavy lifting happens here
		run_ids = self.run_tasks(ctx, job, jtasks, schema_descr, logger)

		feedback = f"Job processed (db: {job.mongo.prefix}"
		if len(run_ids) == 1:
			feedback += f", run id: {run_ids[0]})"
		elif len(run_ids) > 1:
			feedback += ", run ids: " + " ".join([str(el) for el in run_ids]) + ")"
		else:
			feedback = ")"

		logger.info(feedback)
		logger.info(f'Time required: {get_time_delta(start_time)}\n')

		if args.get('show_plots') or args.get('show_plots_cmd'):

			cmd = [
				'ampel', 'plot', 'show', '-stack', '300',
				'-png', os.environ.get('AMPEL_PLOT_DPI', '150'),
				'-t2', '-t3', '-base-path', 'body.plot', # to be improved later
				'-db', job.mongo.prefix, '-run-id', *[str(el) for el in run_ids],
			]

			if args.get('debug'):
				cmd.append('-debug')

			print(
				f"{'-'*40}\n{'Executing' if args.get('show_plots') else 'Plot'} command: {' '.join(cmd)}"
			)

			if args.get('show_plots'):
				rr = subprocess.run(cmd, check=True, capture_output=True, text=True)
				print(rr.stdout)
				print(rr.stderr)

		if args['print_schema_after']:
			print('\nJob schema:\n')
			print('#'*50)
			print(
				'\n' + yaml.dump(
					job.dict(exclude_unset=True),
					sort_keys=False, default_flow_style=None
				)
			)


	def run_tasks(self,
		ctx: AmpelContext,
		job: JobModel,
		jtasks: list[dict[str, Any]],
		schema_descr: str,
		logger: AmpelLogger
	) -> list[int]:

		run_ids = []
		for i, taskd in enumerate(jtasks):

			process_name = f'{job.name or schema_descr}#{i}'

			if isinstance(taskd.get('template', None), dict) and 'live' in taskd['template']:
				taskd = apply_templates(ctx, taskd['template']['live'], taskd, logger)  # noqa: PLW2901
				del taskd['template']

			if 'title' in taskd:
				self.print_chapter(taskd['title'] if taskd.get('title') else f'Task #{i}', logger)
				#process_name += f' [{taskd['title']}]'
				del taskd['title']
			elif i != 0:
				self.print_chapter(f'Task #{i}', logger)
				
			if (multiplier := taskd.pop('multiplier', 1)) > 1:

				try:

					process_queues: list[Process] = []
					result_queues: list[Any] = []

					signal.signal(signal.SIGINT, signal_handler)
					signal.signal(signal.SIGTERM, signal_handler)

					for _ in range(multiplier):
						result_queue: Queue = Queue()
						p = Process(
							target = run_mp_process,
							args = (result_queue, ctx.config._config, taskd, process_name),  # noqa: SLF001
							daemon = True
						)
						p.start()
						process_queues.append(p)
						result_queues.append(result_queue)
					
					for replica, (p, r1) in enumerate(zip(process_queues, result_queues, strict=False)):
						p.join()
						if (m := r1.get()):
							logger.info(f'{taskd["unit"]}#{replica} return value: {m}')

				except KeyboardInterrupt:
					sys.exit(1)
			
			else:
				
				proc = ctx.loader.new_context_unit(
					model = UnitModel(**taskd),
					context = ctx,
					process_name = process_name,
					job_sig = job.sig,
					sub_type = AbsEventUnit,
					base_log_flag = LogFlag.MANUAL_RUN
				)

				event_hdlr = EventHandler(
					proc.process_name,
					ctx.get_database(),
					raise_exc = proc.raise_exc,
					job_sig = job.sig,
					extra = {'task': i}
				)

				x = proc.run(event_hdlr)
				if event_hdlr.run_id:
					run_ids.append(event_hdlr.run_id)

				logger.info(f'{taskd["unit"]} return value: {x}')

		return run_ids


	def load_tasks(self,
		ctx: AmpelContext, job: JobModel, logger: AmpelLogger, debug: bool = False
	) -> list[dict[str, Any]]:

		jtasks: list[dict[str, Any]] = []
		for i, model in enumerate(job.task):
			if isinstance(model.template, Sequence):
				taskd = apply_templates(ctx, model.template, model.dict(exclude_unset=True), logger)
				taskd.pop('template', None)
				if debug:
					self.print_task(taskd, logger)
			elif (
				isinstance(model.template, dict) and
				'pre' in model.template and model.template['pre']
			):
				taskd = apply_templates(ctx, model.template['pre'], model.dict(), logger)
				del taskd['template']['pre']
				if debug:
					self.print_task(taskd, logger)
			else:
				taskd = self.get_task_dict(model)

			taskd['override'] = taskd.pop('override', {}) | {'raise_exc': True}
			# Beacons have no real use in jobs (unlike prod)
			if taskd['unit'] == 'T2Worker' and 'send_beacon' not in taskd['config']:
				taskd['config']['send_beacon'] = False

			jtasks.append(taskd)
			logger.info(
				f'Registering job task#{i} with {model.get_multiplier()}x multiplier'
			)

		return jtasks


	def print_task(self, taskd: dict[str, Any], logger: AmpelLogger) -> None:
		logger.info('Task model morphed by template:')
		for el in prettyjson(taskd, indent=4).split('\n'):
			logger.info(el)


	def get_task_dict(self, task_model: JobTaskModel) -> dict[str, Any]:
		return task_model.dict(exclude_unset=True)


	def get_job_dict(self, job: JobModel, jtasks: list[dict[str, Any]]) -> dict[str, Any]:
		out = job.dict(exclude_unset=True)
		out['task'] = jtasks
		return out


	@staticmethod
	def print_chapter(msg: str, logger: AmpelLogger) -> None:
		logger.info(' ')
		logger.info('=' * (space := (len(msg) + 4)))
		logger.info('‖ ' + msg + ' ‖')
		logger.info('=' * space)
		logger.info(' ')


	@classmethod
	def get_job_schema(cls,
		schema_paths: None | Sequence[str] = None,
		schema_content: None | str = None,
		compute_sig: bool = True,
		Model: type = JobModel
	) -> tuple[None, None] | tuple[JobModel, int]:

		if not (schema_paths or schema_content):
			raise ValueError("Please provide either job file path(s) or content")

		if schema_paths:
			content = io.StringIO()
			for _, job_fname in enumerate(schema_paths):

				if not os.path.exists(job_fname):
					with out_stack():
						raise FileNotFoundError(f'Job file not found: "{job_fname}"\n')

				with open(job_fname) as f:
					content.write('\n'.join(f.readlines()))

			content.seek(0)

		job = yaml.safe_load(schema_content or content)

		if not job:
			return None, None

		for k in list(job.keys()):
			# job root keys starting with % are used by own convention for yaml anchors
			# and thus need not be included in the loaded job structure
			if k.startswith('%'):
				del job[k]

		return Model(**job), build_unsafe_dict_id(job, size=-64) if compute_sig else 0


	@staticmethod
	def _patch_config(config_dict: dict[str, Any], job: JobModel, logger: AmpelLogger):

		# Add channel(s)
		for chan in job.channel:
			logger.info(f'Registering job channel "{chan.channel}"')
			dict.__setitem__(config_dict['channel'], chan.channel, chan.dict())

		# Add aliase(s)
		for k, v in job.alias.items():
			if 'alias' not in config_dict:
				dict.__setitem__(config_dict, 'alias', {})
			for kk, vv in v.items():
				logger.info(f'Registering job alias "{kk}"')
				if k not in config_dict['alias']:
					dict.__setitem__(config_dict['alias'], k, {})
				dict.__setitem__(config_dict['alias'][k], kk, vv)


def run_mp_process(
	result_queue: Queue,
	config: dict[str, Any],
	tast_unit_model: dict[str, Any],
	process_name: str,
	job_sig: None | int = None,
	task_nbr: None | int = None,
	log_profile: str = 'default'
) -> None:

	try:

		# Create new context with serialized config
		context = DevAmpelContext.load(config, one_db=True)

		processor = context.loader.new_context_unit(
			model = UnitModel(**tast_unit_model),
			context = context,
			sub_type = AbsEventUnit,
			log_profile = log_profile,
			process_name = process_name,
			job_sig = job_sig
		)

		eh = EventHandler(
			processor.process_name,
			context.get_database(),
			raise_exc = processor.raise_exc,
			job_sig = job_sig,
			extra = {'task': task_nbr}
		)

		result_queue.put(
			processor.run(eh)
		)

	except Exception as e:
		import traceback
		se = str(e)
		result_queue.put(
			'\n' + '#'*len(se) + '\n' + str(e) + '\n' + '#'*len(se) + '\n' +
			''.join(traceback.format_exception(type(e), e, e.__traceback__))
		)


def signal_handler(sig, frame):
	#import traceback
	print('Interrupt detected')
	#print('Stack frames:')
	#traceback.print_stack(frame)
	raise KeyboardInterrupt()


def edit_job(fname: str) -> str:
	"""
	:raises ValueError if data edition fails
	"""
	if subprocess.call(
		os.environ.get('EDITOR', 'vi') + ' ' + fname,
		shell=True
	):
		with out_stack():
			raise ValueError("Cancellation requested during schema edition")

	return fname



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/logging/LogFlag.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.12.2017
# Last Modified Date:  11.03.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from enum import IntFlag


# flake8: noqa: E221
class LogFlag(IntFlag):
	"""
	Flag used for each log entry stored in the DB.
	Value fits in a MongoDB int32.
	bits 0-5: tier
	bits 5-7: run type
	bits 7-8: location (core system or base units)
	bits 9-14: log level
	"""

	# Execution layer
	T0                      = 1
	T1                      = 2
	T2                      = 4
	T3                      = 8
	T4                      = 16

	# Run type
	SCHEDULED_RUN           = 32
	MANUAL_RUN              = 64

	# Location
	UNIT                    = 128
	CORE                    = 256

	# Log level
	DEBUG                   = 1<<9
	VERBOSE                 = 1<<10
	INFO                    = 1<<11
	SHOUT                   = 1<<12 # SHOUT is for convenience only, saved as INFO into DB
	WARNING                 = 1<<13
	ERROR                   = 1<<14



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t1/T1SimpleRetroCombiner.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                25.05.2021
# Last Modified Date:  23.07.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Generator, Iterable

from ampel.abstract.AbsT1CombineUnit import AbsT1CombineUnit
from ampel.content.DataPoint import DataPoint
from ampel.struct.T1CombineResult import T1CombineResult
from ampel.types import DataPointId


class T1SimpleRetroCombiner(AbsT1CombineUnit):
	"""
	[el.payload for el in combine([{'id': 7}, {'id': 6}, {'id': 5}])]
	will return:
	[[7, 6, 5], [6, 5], [5]]
	"""

	def combine(self, datapoints: Iterable[DataPoint]) -> list[T1CombineResult]: # type: ignore[override]
		"""
		:param datapoints: dict instances representing datapoints
		"""

		chan = self.channel
		dps = [
			dp for dp in datapoints
			if not("excl" in dp and chan in dp['excl'])
		]

		return [T1CombineResult(dps=el) for el in reversed(list(self.generate_retro_sequences(dps)))]

	def generate_retro_sequences(self, datapoints: list[DataPoint]) -> Generator[list[DataPointId], None, None]:
		"""
		Generate substates by iteratively removing the last element. This may
		be overridden by subclasses, e.g. to use only certain datapionts to
		delimit states.
		"""
		while datapoints:
			yield [dp["id"] for dp in datapoints]
			datapoints = datapoints[:-1]




from collections.abc import Iterable
from typing import Generic

from ampel.base.AmpelABC import AmpelABC
from ampel.base.AmpelUnit import AmpelUnit
from ampel.base.decorator import abstractmethod
from ampel.types import (
	T,
)


class AbsConsumer(AmpelABC, AmpelUnit, Generic[T], abstract=True):

	@abstractmethod
	def consume(self) -> None | T:
		"""Get a single message from the queue"""
		...
	
	@abstractmethod
	def acknowledge(self, docs: Iterable[T]) -> None:
		"""Acknowledge the processing of a batch of messages"""
		...


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/util/distrib.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                13.03.2021
# Last Modified Date:  23.04.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import os
import re
from collections.abc import Generator, MutableSequence, Sequence
from importlib import metadata
from pathlib import Path
from typing import TypeAlias

# NB: PackagePath implements read_text(), but is not a subclass of Path
PathLike: TypeAlias = Path | metadata.PackagePath
PathList: TypeAlias = MutableSequence[PathLike]

def get_dist_names(distrib_prefixes: Sequence[str] = ("ampel-", "pyampel-")) -> list[str]:
	""" Get all installed distributions whose names start with the provided prefix """
	# ensure that at least interface and core are found
	prefixes = {"ampel-interface", "ampel-core"}.union(distrib_prefixes)
	ret = [
		dist.name for dist in metadata.distributions()
		if any(prefix in dist.name for prefix in prefixes)
	]

	if ret:
		ret.insert(0, ret.pop(next(i for i, el in enumerate(ret) if "interface" in el)))
		ret.insert(0, ret.pop(next(i for i, el in enumerate(ret) if "core" in el)))

	return ret


def get_files(
	dist_name: str,
	lookup_dir: None | str = None,
	pattern: None | re.Pattern = None
) -> PathList:
	""" Loads all known conf files of the provided distribution (name) """

	if lookup_dir and not lookup_dir.endswith("/"):
		lookup_dir += "/"

	files: PathList = []

	for p in metadata.files(dist_name) or []:
		# Path config file from editable install: look for files in the
		# referenced path(s)
		if p.suffix == ".pth":
			for path in p.read_text().splitlines():
				if os.path.isdir(path):
					files.extend(walk_dir(
						path,
						lookup_dir,
						pattern
					))
				else:
					# setuptools editable installs work by installing a path
					# hook. There's no way to get at the actual path to the
					# project without depending deeply on assumptions about how
					# setuptools currently works, so just bail.
					raise ValueError(
						f"Invalid path '{path}' in {p}. If {dist_name} was "
						"installed with setuptools, it must be installed "
						"normally (not in editable mode)."
					)
		# Wheel installs have the files directly in the distribution
		elif _check_match(p.as_posix(), lookup_dir, pattern):
			files.append(p)
	return files


def _check_match(
	arg: str,
	lookup_dir: None | str,
	pattern: None | re.Pattern
) -> bool:
	if lookup_dir and not arg.startswith(lookup_dir):
		return False
	return not (pattern and not pattern.match(arg))


def walk_dir(
	path: str,
	lookup_dir: None | str = None,
	pattern: None | re.Pattern = None
) -> Generator[Path, None, None]:
	for root, _, files in os.walk(f"{path}/{lookup_dir}" if lookup_dir else f"{path}"):
		for fname in files:
			if pattern and not pattern.match(fname):
				continue
			yield Path(root, fname)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/core/AmpelController.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.04.2020
# Last Modified Date:  14.03.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import asyncio
import re
from collections.abc import Sequence
from typing import TYPE_CHECKING, Literal

from ampel.abstract.AbsProcessController import AbsProcessController
from ampel.config.AmpelConfig import AmpelConfig
from ampel.core.UnitLoader import UnitLoader
from ampel.log.AmpelLogger import DEBUG, VERBOSE, AmpelLogger
from ampel.model.ProcessModel import ProcessModel
from ampel.secret.AmpelVault import AmpelVault
from ampel.util.hash import build_unsafe_dict_id

if TYPE_CHECKING:
	from ampel.protocol.LoggerProtocol import LoggerProtocol


class AmpelController:
	"""
	Top-level controller class that spawns "process controllers" (subclasses of AbsProcessController).
	An ampel config instance or a path to an ampel config is required.
	The process definitions therein (listed under key 'process') are parsed and
	used to spawn the corresponding process controllers.
	The parameters 'tier', 'match', 'exclude' and 'controllers' can be used to sub-select processes.
	"""

	def __init__(
		self,
		config_arg: str | AmpelConfig,
		vault: None | AmpelVault = None,
		tier: None | Literal[0, 1, 2, 3] = None,
		match: None | Sequence[str] = None,
		exclude: None | Sequence[str] = None,
		controllers: None | Sequence[str] = None,
		logger: 'None | LoggerProtocol' = None,
		verbose: int = 0,
		**kwargs,
	):
		"""
		:param config_arg: ampel config or path to ampel config file (json or yaml)
		:param pwd_file_path: if provided, the encrypted conf entries possibly contained \
			in the ampel config instance will be decrypted using the provided password file. \
			The password file must define one password per line.
		:param tier: if specified, only processes defined under the given tier will be scheduled
		:param match: list of regex strings to be matched against process names.
			Only matching processes will be scheduled.
		:param exclude: list of regex strings to be matched against process names.
			Only non-matching processes will be scheduled.
		:param controllers: list of controller class names to be matched against process controller definitions.
			Only processes with matching controller units will be scheduled.
		:param logger: if not provided, a new logger will be instantiated using AmpelLogger.get_logger()
		:param verbose: 1 -> verbose, 2 -> debug
		:param kwargs: will be forwared to the constructor of ampel process controllers
		"""

		self.controllers: list[AbsProcessController] = []
		if isinstance(config_arg, str):
			config = AmpelConfig.load(config_arg, freeze=False)
		else:
			config = config_arg
		loader = UnitLoader(config, db=None, provenance=False, vault=vault)

		if verbose:
			if not logger:
				logger = AmpelLogger.get_logger(
					console={"level": DEBUG if verbose > 1 else VERBOSE}
				)
			logger.log(VERBOSE, "Config file loaded")

		for processes in self.group_processes(
			self.get_processes(
				config,
				tier=tier,
				match=match,
				exclude=exclude,
				controllers=controllers,
				logger=logger,
				verbose=verbose,
			)
		):
			if logger and verbose:
				logger.log(
					VERBOSE,
					f"Spawing new {processes[0].controller.unit} with processes: {list(p.name for p in processes)}",
				)
			controller_kwargs = {
				"config": config,
				"vault": vault,
				"processes": processes,
				**kwargs,
			}

			self.controllers.append(
				loader.new(
					processes[0].controller,
					unit_type=AbsProcessController,
					**controller_kwargs,
				)
			)


	async def run(self):

		tasks = [
			asyncio.create_task(controller.run())
			for controller in self.controllers
		]
		task = asyncio.gather(*tasks, return_exceptions=True)

		try:
			return await task
		except asyncio.CancelledError:
			for t in tasks:
				t.cancel()
			return await task


	@staticmethod
	def group_processes(processes: list[ProcessModel]) -> list[list[ProcessModel]]:
		"""
		Group processes by controller
		"""
		d: dict[int, list[ProcessModel]] = {}
		for pm in processes:
			controller_id = build_unsafe_dict_id(
				pm.controller.dict(exclude_unset=True), ret=int
			)
			if controller_id in d:
				# Gather process (might raise error in case of invalid process)
				d[controller_id].append(pm)
				continue
			d[controller_id] = [pm]
		return [v for v in d.values()]


	@staticmethod
	def get_processes(
		config: AmpelConfig,
		tier: None | Literal[0, 1, 2, 3, "ops"] = None,
		match: None | Sequence[str] = None,
		exclude: None | Sequence[str] = None,
		controllers: None | Sequence[str] = None,
		logger: 'None | LoggerProtocol' = None,
		verbose: int = 0,
		raise_exc: bool = False,
	) -> list[ProcessModel]:
		"""
		Extract processes from the config. Only active processes are returned.

		:param tier: if specified, only processes defined under a given tier will be returned
		:param match: list of regex strings to be matched against process names.
			Only matching processes will be returned
		:param exclude: list of regex strings to be matched against process names.
			Only non-matching processes will be returned
		:param logger: if provided, information about ignored/excluded processes will be logged
		:param verbose: 1 -> verbose, 2 -> debug
		:param raise_exc: if True, raise ValidationError on invalid processes
		"""

		ret: list[ProcessModel] = []

		if match:
			rmatch = [re.compile(el) for el in match]  # Compile regexes

		if exclude:
			rexcl = [re.compile(el) for el in exclude]

		for t in [tier] if tier is not None else [0, 1, 2, 3, "ops"]:

			tier_name = f"t{t}" if isinstance(t, int) else t
			for p in config.get(f"process.{tier_name}", dict, raise_exc=True).values():

				# Process name inclusion filter
				if match and not any(rm.match(p["name"]) for rm in rmatch):
					if logger and verbose > 1:
						logger.debug(
							f'Ignoring process {p["name"]} unmatched by {rmatch}'
						)
					continue

				# Process name exclusion filter
				if exclude and any(rx.match(p["name"]) for rx in rexcl):
					if logger and verbose > 1:
						logger.info(
							f'Excluding process {p["name"]} matched by {rmatch}'
						)
					continue

				if not p.get("active", True):
					if logger:
						logger.log(VERBOSE, f"Ignoring inactive process {p.get('name')}")
					continue

				try:
					# Set defaults
					pm = ProcessModel(**p)
				except Exception as e:
					if logger:
						logger.error(f"Unable to load invalid process {p}", exc_info=e)
					if raise_exc:
						raise
					continue

				# Controller exclusion
				if controllers and pm.controller.unit not in controllers:
					if logger:
						logger.log(
							VERBOSE,
							f"Ignoring process {pm.name} with controller {pm.controller.unit}",
						)
					continue

				ret.append(pm)

		return ret


	@classmethod
	def main(cls, args: None | list[str] = None) -> None:

		import logging
		import signal
		from argparse import ArgumentParser

		from ampel.secret.DictSecretProvider import DictSecretProvider

		logging.basicConfig(level="INFO")

		def maybe_int(stringy):
			try:
				return int(stringy)
			except Exception:
				return stringy

		parser = ArgumentParser(description="""Run Ampel processes""")
		parser.add_argument("config_arg", help="Path to Ampel config file")
		parser.add_argument(
			"--secrets",
			default=None,
			help="Path to a YAML secrets store in sops format",
		)
		parser.add_argument(
			"--tier",
			type=maybe_int,
			choices=(0, 1, 2, 3, "ops"),
			default=None,
			help="Run only processes of this tier",
		)
		parser.add_argument(
			"--match",
			type=str,
			nargs="*",
			default=None,
			help="Run only processes whose names match at least one of these regexes",
		)
		parser.add_argument("-v", "--verbose", action="count", default=0)
		args_ns = parser.parse_args(args)
		secrets_file = args_ns.secrets
		if secrets_file is not None:
			args_ns.secrets = DictSecretProvider.load(secrets_file)

		mcp = cls(**args_ns.__dict__)

		def handle_signals(task, loop, graceful=True):
			for s in (signal.SIGHUP, signal.SIGTERM, signal.SIGINT):
				loop.remove_signal_handler(s)
				loop.add_signal_handler(
					s,
					lambda s=s, task=task, loop=loop, graceful=graceful: asyncio.create_task(
						shutdown(s, task, loop, graceful)
					),
				)

		async def shutdown(sig, task, loop, graceful=True):
			"""Stop root task on signal"""
			if graceful:
				logging.info(
					f"Received exit signal {sig.name}, shutting down gracefully (signal again to terminate immediately)..."
				)
				for controller in mcp.controllers:
					controller.stop()
				handle_signals(task, loop, False)
			else:
				logging.info(
					f"Received exit signal {sig.name}, terminating immediately..."
				)
				task.cancel()
			await task
			loop.stop()

		def reload_config() -> None:

			try:
				logging.info(f"Reloading config from {args_ns.config_arg}")
				config = AmpelConfig.load(args_ns.config_arg, freeze=False)
				loader = UnitLoader(
					config,
					db=None,
					provenance=False,
					vault=AmpelVault(
						providers=[DictSecretProvider.load(secrets_file)] 
					) if secrets_file else None,
				)
				# Ensure that process models are valid
				with loader.validate_unit_models():
					groups = AmpelController.group_processes(
						AmpelController.get_processes(
							config, tier=args_ns.tier, match=args_ns.match,
						)
					)
			except Exception:
				logging.exception(f"Failed to load {args_ns.config_arg}")
				return

			try:
				controllers = list(mcp.controllers)
				matches = []
				for group in groups:
					names = {pm.name for pm in group}
					for i, candidate in enumerate(controllers):
						if names.intersection([pm.name for pm in candidate.processes]):
							matches.append((candidate, group))
							del controllers[i]
							break
					else:
						raise RuntimeError(f"No match for process group {names}")
				assert len(matches) == len(mcp.controllers)
			except Exception:
				logging.exception("Failed to match process groups with current set")

			for controller, processes in matches:
				try:
					controller.update(config, loader.vault, processes)
					logging.info(
						f"Updated {controller.__class__.__name__} with processes: {[pm.name for pm in processes]} "
					)
				except Exception:  # noqa: PERF203
					logging.exception(
						f"Failed to update {controller.__class__.__name__} with processes: {[pm.name for pm in processes]}"
					)

		loop = asyncio.get_event_loop()
		task = loop.create_task(mcp.run())
		handle_signals(task, loop)
		loop.add_signal_handler(signal.SIGUSR1, reload_config)

		for result in loop.run_until_complete(task):
			if isinstance(result, asyncio.CancelledError):
				...
			elif isinstance(result, BaseException):
				raise result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/time/QueryTimeModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                10.12.2019
# Last Modified Date:  06.06.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any

from ampel.base.AmpelBaseModel import AmpelBaseModel


class QueryTimeModel(AmpelBaseModel):
	""" Standardized parameter for class QueryMatchStock """

	before: None | int | float = None
	after: None | int | float = None

	def __bool__(self) -> bool:
		return self.before is not None or self.after is not None

	def dict(self, **kwargs) -> dict[str, Any]:
		d = super().dict(**kwargs)
		d['$lt'] = d.pop('before')
		d['$gt'] = d.pop('after')
		return {k: v for k, v in d.items() if v is not None}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsT4ControlUnit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                03.04.2023
# Last Modified Date:  04.04.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Generator

from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.content.T4Document import T4Document
from ampel.core.ContextUnit import ContextUnit
from ampel.core.EventHandler import EventHandler
from ampel.log.AmpelLogger import AmpelLogger
from ampel.types import ChannelId, OneOrMany, Traceless


class AbsT4ControlUnit(ContextUnit, AmpelABC, abstract=True):

	logger: Traceless[AmpelLogger]
	event_hdlr: Traceless[EventHandler]
	channel: None | OneOrMany[ChannelId] = None
	
	@abstractmethod
	def do(self) -> Generator[T4Document, None, None]:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/query/general.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                11.12.2019
# Last Modified Date:  17.02.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any, Literal

from bson.int64 import Int64

from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.OneOf import OneOf
from ampel.mongo.schema import apply_excl_schema, apply_schema
from ampel.mongo.utils import maybe_match_array
from ampel.types import ChannelId, StockId, StrictIterable, Tag

type_stock_id = (int, Int64, bytes, str)


def build_general_query(
	stock: None | StockId | StrictIterable[StockId] = None,
	channel: None | ChannelId | dict | AllOf[ChannelId] | AnyOf[ChannelId] | OneOf[ChannelId] = None,
	tag: None | dict[Literal['with', 'without'], Tag | dict | AllOf[Tag] | AnyOf[Tag] | OneOf[Tag]] = None
) -> dict[str, Any]:
	"""
	Builds a query usable with the ampel "stock", "t0" (with channel=None), "t1" and "t2" collections
	:param stock: matching multiple ids with a single query is possible
	:param channel: None (no criterium) means all channel are considered.
	:param tag: tags to be (or not to be) matched by query
	:returns: query dict with matching criteria
	:raises ValueError: apply_schema can raise ValueError in case the provided dict schema structure is unsupported
	"""

	query = {}

	if stock:
		query['stock'] = stock if isinstance(stock, type_stock_id) \
			else maybe_match_array(stock)

	if channel:
		apply_schema(query, 'channel', channel)

	if tag:

		if 'with' in tag:
			apply_schema(query, 'tag', tag['with'])

		# Order matters, parse_dict(...) must be called *after* parse_excl_dict(...)
		if 'without' in tag:
			apply_excl_schema(query, 'tag', tag['without'])

	return query



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/update/MongoT1Ingester.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                24.04.2021
# Last Modified Date:  09.10.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any

from pymongo import UpdateOne

from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.content.T1Document import T1Document
from ampel.mongo.update.HasUpdatesBuffer import HasUpdatesBuffer
from ampel.mongo.utils import maybe_use_each


class MongoT1Ingester(AbsDocIngester[T1Document], HasUpdatesBuffer):

	def ingest(self, doc: T1Document) -> None:

		# Note: $setOnInsert does not retain key order
		set_on_insert: dict[str, Any] = {'dps': doc['dps']}
		add_to_set: dict[str, Any] = {'channel': maybe_use_each(doc['channel'])}

		match: dict[str, Any] = {
			'stock': doc['stock'],
			'link': doc['link']
		}

		if 'unit' in doc:
			set_on_insert['unit'] = doc['unit']
			match['unit'] = doc['unit']

		if 'config' in doc:
			match['config'] = doc['config']

		if 'origin' in doc:
			match['origin'] = doc['origin']

		if 'tag' in doc:
			add_to_set['tag'] = maybe_use_each(doc['tag'])

		if 'body' in doc:
			set_on_insert['body'] = doc['body']

		self.updates_buffer.add_t1_update(
			UpdateOne(
				match,
				{
					'$setOnInsert': set_on_insert,
					'$addToSet': add_to_set,
					'$max': {'expiry': doc['expiry']} if 'expiry' in doc else {},
					 # meta must be set by compiler
					'$push': {'meta': maybe_use_each(doc['meta'])},
				},
				upsert=True
			)
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/stage/project/AbsT3Projector.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                21.01.2020
# Last Modified Date:  21.04.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Iterable, Sequence

from ampel.base.AmpelABC import AmpelABC
from ampel.base.AmpelUnit import AmpelUnit
from ampel.base.decorator import abstractmethod
from ampel.struct.AmpelBuffer import AmpelBuffer


class AbsT3Projector(AmpelABC, AmpelUnit, abstract=True):

	@abstractmethod
	def project(self, seq: Iterable[AmpelBuffer]) -> Sequence[AmpelBuffer]:
		...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/job/EnvSpec.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                Unspecified
# Last Modified Date:  23.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.base.AmpelBaseModel import AmpelBaseModel


class EnvSpec(AmpelBaseModel):
    set: dict[str, str] = {}
    check: dict[str, int|str] = {}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/t3/T3ProjectionDirective.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                02.03.2021
# Last Modified Date:  02.03.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>


from collections.abc import Sequence

from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.model.UnitModel import UnitModel


class T3ProjectionDirective(AmpelBaseModel):
	"""
	Internal model used for field 'directives' of T3UnitRunner
	"""

	#: AbsT3Filter sub unit to use for down-selection of ampel buffers
	filter: None | UnitModel

	#: AbsT3Projector sub unit capable of discarding selected ampel buffer attributes/fields
	project: None | UnitModel

	#: t3 units (AbsT3Unit) to execute
	execute: Sequence[UnitModel]

	def __init__(self, **kwargs):
		if isinstance(v := kwargs.get("execute"), dict | UnitModel):
			kwargs["execute"] = [v]
		super().__init__(**kwargs)



from contextlib import nullcontext

import pytest
import schedule as sched

from ampel.config.ScheduleEvaluator import ScheduleEvaluator


@pytest.mark.parametrize(
    ("line","unit","interval","error"),
    [
        ("every(30).seconds", "seconds", 30, None),
        ("every().day.at('16:00')", "days", 1, None),
        ("requests.get('http://malicio.us')", None, None, ValueError),
    ],
)
def test_evaluate(line, unit, interval, error):
    evaluator = ScheduleEvaluator()
    with pytest.raises(error) if error else nullcontext():
        job = evaluator(sched.Scheduler(), line)
        assert job.unit == unit
        assert job.interval == interval



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/ViewCommand.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                15.03.2021
# Last Modified Date:  27.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from argparse import ArgumentParser
from collections.abc import Sequence
from typing import Any

from ampel.cli.AbsLoadCommand import AbsLoadCommand
from ampel.cli.AbsStockCommand import AbsStockCommand
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser
from ampel.cli.ArgParserBuilder import ArgParserBuilder
from ampel.cli.utils import maybe_load_idmapper
from ampel.core.AmpelContext import AmpelContext
from ampel.log.LogFlag import LogFlag
from ampel.model.t3.T3ProjectionDirective import T3ProjectionDirective
from ampel.model.UnitModel import UnitModel
from ampel.t3.T3Processor import T3Processor

hlp = {
	'config': 'Path to an ampel config file (yaml/json)',
	# Optional
	'secrets': 'Path to a YAML secrets store in sops format',
	'db': 'Database prefix. If set, "-mongo.prefix" value will be ignored',
	'out': 'Path to file where serialized views will be saved (printed to stdout otherwise)',
	'log-profile': 'One of: default, compact, headerless, verbose, debug',
	'id-mapper': 'Convert stock ids using the provided id mapper (ex: ZTFIdMapper)',
	'channel': 'view for specified'
}

class ViewCommand(AbsStockCommand, AbsLoadCommand):
	"""
	TODO: implement
	"""

	@staticmethod
	def get_sub_ops() -> list[str]:
		return ['show', 'save', 'feed']

	# Mandatory implementation
	def get_parser(self, sub_op: None | str = None) -> ArgumentParser | AmpelArgumentParser:

		if sub_op in self.parsers:
			return self.parsers[sub_op]

		sub_ops = self.get_sub_ops()
		if sub_op is None or sub_op not in sub_ops:
			return AmpelArgumentParser.build_choice_help(
				't2', sub_ops, hlp,
				description = 'Create and show or save selected ampel views.\n' +
				'These can also be provided to T3 units using the operation feed.'
			)

		builder = ArgParserBuilder('view')
		hlp.update(self.get_select_args_help())
		hlp.update(self.get_load_args_help())
		builder.add_parsers(sub_ops, hlp)

		# Required args
		builder.req('out', 'save', default=True)
		builder.req('config')

		# Optional args
		builder.opt('secrets', default=None)
		builder.opt('id-mapper')
		builder.opt('debug', action='store_true')
		builder.opt('log-profile', default='default')
		builder.opt('binary', action='store_true')

		# Selection args
		self.add_selection_args(builder)

		# Content args
		self.add_load_args(builder, 'Views content arguments')

		# Example
		builder.example('show', 'view show -channel CHAN1 -stock 85628462 -no-t0')
		builder.example('save', '-stock 519059889 -db MyDB -no-t0 -channel CHAN1 -out file.txt -binary')
		
		self.parsers.update(
			builder.get()
		)

		return self.parsers[sub_op]


	# Mandatory implementation
	def run(self, args: dict[str, Any], unknown_args: Sequence[str], sub_op: None | str = None) -> None:

		if not args['channel']:
			print('Option "channel" is required')
			return

		ctx = self.get_context(
			args, unknown_args, ContextClass = AmpelContext,
			require_existing_db = args['db'] or True, one_db='auto'
		)

		maybe_load_idmapper(args)

		conf = {
			'fd': args.get('out'),
			'id_mapper': args['id_mapper'],
			'verbose': sub_op == 'save',
			'binary': args.get('binary') or False
		}

		t3p = T3Processor(
			context = ctx,
			process_name = 'ViewCommand',
			template = 'compact_t3',
			base_log_flag = LogFlag.MANUAL_RUN,
			log_profile = 'console_debug' if args.get('debug') else 'console_info',
			execute = [
				{
					'supply': {
						'unit': 'T3DefaultBufferSupplier',
						'config': {
							'select': self.build_select_model(args),
							'load': self.build_load_model(args)
						}
					},
					'stage': {
						'unit': 'T3ProjectingStager',
						'config': {
							'keep_buffers': True, # quick n dirty
							'directives': [
								T3ProjectionDirective(
									project=UnitModel(
										unit='T3ChannelProjector',
										config = {'channel': args['channel']}
									),
									execute=[
										UnitModel(
											unit='T3BufferExporterUnit',
											config=conf
										)
									],
								),
							],
						}
					}
				}
			]
		)

		t3p.run()



from pymongo import InsertOne

from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry
from ampel.mongo.update.DBUpdatesBuffer import DBUpdatesBuffer


def test_metrics(dev_context, ampel_logger):
    updates_buffer = DBUpdatesBuffer(dev_context.db, run_id=0, logger=ampel_logger)
    get_sample_value = AmpelMetricsRegistry.registry().get_sample_value
    cols = ["stock", "t0", "t1", "t2"]

    before = {k: get_sample_value("ampel_db_ops_total", {"col": k}) or 0 for k in cols}
    updates_buffer.add_t0_update(InsertOne({"_id": 0, "foo": "bar"}))
    updates_buffer.push_updates()
    after = {k: get_sample_value("ampel_db_ops_total", {"col": k}) or 0 for k in cols}

    for k in cols:
        if k == "t0":
            assert after[k] - before[k] == 1, f"ops count was incremented for {k}"
        else:
            assert after[k] - before[k] == 0, f"ops count was not incremented for {k}"

    before = {k: get_sample_value("ampel_db_errors_total", {"col": k}) or 0 for k in cols}
    updates_buffer.add_t0_update(InsertOne({"_id": 0, "foo": "bar"}))
    updates_buffer.push_updates()
    after = {k: get_sample_value("ampel_db_errors_total", {"col": k}) or 0 for k in cols}

    for k in cols:
        if k == "t0":
            assert after[k] - before[k] == 1, f"error count was incremented for {k}"
        else:
            assert after[k] - before[k] == 0, f"error count was not incremented for {k}"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsConfigUpdater.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                05.04.2023
# Last Modified Date:  05.04.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any

from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.core.AmpelContext import AmpelContext
from ampel.log.AmpelLogger import AmpelLogger


class AbsConfigUpdater(AmpelABC, abstract=True):
	"""
	Context is required as alteration products might need to be registered in context sub-elements.
	Note that unlike AbsConfigMorpher, this abstract class is not a model (AmpelBaseModel).
	Implementing classes ex: HashT2Config, ResolveRunTimeAliases.
	"""

	@abstractmethod
	def alter(self, context: AmpelContext, target: dict[str, Any], logger: AmpelLogger) -> dict[str, Any]:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/ingest/FilterModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.10.2019
# Last Modified Date:  20.05.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Literal

from ampel.model.UnitModel import UnitModel


class FilterModel(UnitModel[str]):

	#: How to store rejection records
	reject: None | dict[Literal['log', 'register'], UnitModel] = None

	#: How to handle the defined filter if the associated stock is already in the database
	on_stock_match: None | Literal['bypass', 'overrule', 'silent_overrule'] = None
